'''
Created on 20 sep. 2011

.. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>
'''
import numpy as np
import matplotlib.pyplot as plt

from ema_workbench import load_results, ema_logging

from ema_workbench.analysis.pairs_plotting import (pairs_lines, pairs_scatter, 
                                                   pairs_density)

ema_logging.log_to_stderr(level=ema_logging.DEFAULT_LEVEL)

# load the data
fh = './data/1000 flu cases no policy.tar.gz'
experiments, outcomes = load_results(fh)

# transform the results to the required format
# that is, we want to know the max peak and the casualties at the end of the 
# run
tr = {}

# get time and remove it from the dict
time = outcomes.pop('TIME')

for key, value in outcomes.items():
    if key == 'deceased population region 1':
        tr[key] = value[:,-1] #we want the end value
    else:
        # we want the maximum value of the peak
        max_peak = np.max(value, axis=1) 
        tr['max peak'] = max_peak
        
        # we want the time at which the maximum occurred
        # the code here is a bit obscure, I don't know why the transpose 
        # of value is needed. This however does produce the appropriate results
        logical = value.T==np.max(value, axis=1)
        tr['time of max'] = time[logical.T]
        
pairs_scatter((experiments, tr), filter_scalar=False)
pairs_lines((experiments, outcomes))
pairs_density((experiments, tr), filter_scalar=False)
plt.show() 


'''parameters and collections of parameters'''
from __future__ import (unicode_literals, print_function, absolute_import,
                        division)

import abc
import itertools
import numbers
import pandas
import six
import warnings

from ema_workbench.em_framework.util import (NamedObject, Variable,
                                         NamedObjectMap, Counter, NamedDict)

# Created on Jul 14, 2016
#
# .. codeauthor::jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

__all__ = ['Parameter','RealParameter', 'IntegerParameter',
           'CategoricalParameter', 'create_uncertainties', 
           'experiment_generator']


class Constant(NamedObject):
    '''Constant class, 
    
    can be used for any parameter that has to be set to a fixed value
    
    '''
    
    def __init__(self, name, value):
        super(Constant, self).__init__(name)
        self.value = value

    def __repr__(self, *args, **kwargs):
        return '{}(\'{}\', {})'.format(self.__class__.__name__,
                                       self.name, self.value)


class Category(Constant):
    def __init__(self, name, value):
        super(Category, self).__init__(name, value)

def create_category(cat):
    if isinstance(cat, Category):
        return cat
    else:
        return Category(str(cat), cat)
    

class Parameter(Variable):
    ''' Base class for any model input parameter
    
    Parameters
    ----------
    name : str
    lower_bound : int or float
    upper_bound : int or float
    resolution : collection
    pff : bool
          if true, sample over this parameter using resolution in case of
          partial factorial sampling
    
    Raises
    ------
    ValueError 
        if lower bound is larger than upper bound
    ValueError 
        if entries in resolution are outside range of lower_bound and
        upper_bound
    
    '''
    
    __metaclass__ = abc.ABCMeta
        
    INTEGER = 'integer'
    UNIFORM = 'uniform'

    def __init__(self, name, lower_bound, upper_bound, resolution=None,
                 default=None, variable_name=None, pff=False):
        super(Parameter, self).__init__(name)
        
        if resolution is None:
            resolution = []
        
        for entry in resolution:
            if not ((entry >= lower_bound) and (entry <= upper_bound)):
                raise ValueError(('resolution not consistent with lower and ' 
                                  'upper bound'))
        
        if lower_bound >= upper_bound:
            raise ValueError('upper bound should be larger than lower bound')
        
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.resolution = resolution
        self.default = default
        self.variable_name = variable_name
        self.pff = pff
        
    def __eq__ (self, other):
        comparison = [all(hasattr(self, key) == hasattr(other, key) and
                          getattr(self, key) == getattr(other, key) for key 
                          in self.__dict__.keys())]
        comparison.append(self.__class__ == other.__class__)
        return all(comparison)

    def __str__(self):
        return self.name
    
    def __repr__(self, *args, **kwargs):
        start = '{}(\'{}\', {}, {}'.format(self.__class__.__name__, 
                                          self.name, 
                                          self.lower_bound, self.upper_bound)
        
        if self.resolution:
            start += ', resolution={}'.format(self.resolution)
        if self.default:
            start += ', default={}'.format(self.default)
        if self.variable_name != [self.name]:
            start += ', variable_name={}'.format(self.variable_name)
        if self.pff:
            start += ', pff={}'.format(self.pff)
            
        start += ')'
        
        return start
            
            
class RealParameter(Parameter):
    ''' real valued model input parameter
    
    Parameters
    ----------
    name : str
    lower_bound : int or float
    upper_bound : int or float
    resolution : iterable 
    variable_name : str, or list of str
    
    Raises
    ------
    ValueError 
        if lower bound is larger than upper bound
    ValueError 
        if entries in resolution are outside range of lower_bound and
        upper_bound
    
    '''
    
    
    def __init__(self, name, lower_bound, upper_bound, resolution=None, 
                 default=None, variable_name=None, pff=False):
        super(RealParameter, self).__init__(name, lower_bound, upper_bound,
                                    resolution=resolution, default=default,
                                    variable_name=variable_name, pff=pff)
        
        self.dist = Parameter.UNIFORM
        
    @property
    def params(self):
        return (self.lower_bound, self.upper_bound-self.lower_bound)

  
class IntegerParameter(Parameter):
    ''' integer valued model input parameter
    
    Parameters
    ----------
    name : str
    lower_bound : int or float
    upper_bound : int or float
    resolution : iterable
    variable_name : str, or list of str
    
    Raises
    ------
    ValueError 
        if lower bound is larger than upper bound
    ValueError 
        if entries in resolution are outside range of lower_bound and
        upper_bound, or not an numbers.Integral instance
    ValueError 
        if lower_bound or upper_bound is not an numbers.Integral instance
    
    '''
    
    def __init__(self, name, lower_bound, upper_bound, resolution=None, 
                 default=None, variable_name=None, pff=False):
        super(IntegerParameter, self).__init__(name, lower_bound, upper_bound, 
                                    resolution=resolution, default=default,
                                    variable_name=variable_name, pff=pff)
        
        lb_int = isinstance(lower_bound, numbers.Integral) 
        up_int = isinstance(upper_bound, numbers.Integral)
        
        if not (lb_int or up_int):
            raise ValueError('lower bound and upper bound must be integers')
        
        for entry in self.resolution:
            if not isinstance(entry, numbers.Integral):
                raise ValueError(('all entries in resolution should be ' 
                                  'integers'))
        
        self.dist = Parameter.INTEGER
        
    @property
    def params(self):
        return (self.lower_bound, self.upper_bound)
    

class CategoricalParameter(IntegerParameter):
    ''' categorical model input parameter
    
    Parameters
    ----------
    name : str
    categories : collection of obj
    variable_name : str, or list of str
    multivalue : boolean
                 if categories have a set of values, for each variable_name
                 a different one.
    
    '''
    
    @property
    def categories(self):
        return self._categories
    
    @categories.setter
    def categories(self, values):
        self._categories.extend(values)
        
    def __init__(self, name, categories, default=None, variable_name=None, 
                 pff=False, multivalue=False):
        lower_bound = 0
        upper_bound = len(categories)
        
        if upper_bound == 1:
            raise ValueError('there should be more than 1 category')

        super(CategoricalParameter, self).__init__(name, lower_bound, 
                            upper_bound, resolution=None, default=default,
                            variable_name=variable_name, pff=pff)
        cats = [create_category(cat) for cat in categories]
        
        self._categories = NamedObjectMap(Category)
        
        self.categories = cats
        self.resolution = [i for i in range(len(self.categories))]
        self.multivalue = multivalue
        
    def index_for_cat(self, category):
        '''return index of category
        
        Parameters
        ----------
        category : object
        
        Returns
        -------
        int
        
        
        '''
        for i, cat in enumerate(self.categories):
            if cat.name==category:
                return i
        raise ValueError("category not found")

    def cat_for_index(self, index):
        '''return category given index
        
        Parameters
        ----------
        index  : int
        
        Returns
        -------
        object
        
        '''
        
        return self.categories[index]
        
    def transform(self, value):
        '''return index of category
        
        Parameters
        ----------
        name : int
               index for which you want the category
               
        Raises
        ------
        IndexError
            if value is out of bounds
        '''
        warnings.warn('deprecated, use cat_for_index instead')
        return self.cat_for_index(value)
    
    def invert(self, name):
        ''' invert a category to an integer
        
        Parameters
        ----------
        name : obj
               category
               
        Raises
        ------
        ValueError
            if category is not found
        
        '''
        warnings.warn('deprecated, use index_for_cat instead')
        return self.index_for_cat(name)

    def __repr__(self, *args, **kwargs):
        template1 = 'CategoricalParameter(\'{}\', {}, default={})'
        template2 = 'CategoricalParameter(\'{}\', {})'
        
        if self.default:
            representation = template1.format(self.name, self.resolution, 
                                            self.default)
        else:
            representation = template2.format(self.name, self.resolution)
        
        return representation


class Policy(NamedDict):
    # idea: make use of NamedObjectMap --> underneath this is an
    # ordered dict, so can always cast values to list / array
    # makes decisions ala the lake model easier to implement
    # combine this with some kind of function that generates many of the
    # same lever objects but with an _id appended to name
    # and lake problem becomes easy
    #
    # problem: won't work as long as you pass any kwargs using **kwarg
    # notation. This removes the ordering
    # practical issue: switching NamedDict from UserDict to 
    # OrderedDict creates a world of pain
    def __init__(self, name, **kwargs):
        super(Policy, self).__init__(name, **kwargs)
    
    def to_list(self, parameters):
        '''get list like representation of policy where the
        parameters are in the order of levers'''
        
        return [self[param.name] for param in parameters]
        

class Scenario(NamedDict):
    def __init__(self, name=Counter(), **kwargs):
        super(Scenario, self).__init__(name, **kwargs)


class Experiment(NamedObject):
    def __init__(self, name, model_name, policy, scenario, experiment_id):
        super(Experiment, self).__init__(name)
        self.experiment_id = experiment_id
        self.policy = policy
        self.model_name = model_name
        self.scenario = scenario


def experiment_generator(scenarios, model_structures, policies):
    '''
    
    generator function which yields experiments
    
    Parameters
    ----------
    designs : iterable of dicts
    model_structures : list
    policies : list

    Notes
    -----
    this generator is essentially three nested loops: for each model structure,
    for each policy, for each experiment, run the experiment. This means 
    that designs should not be a generator because this will be exhausted after
    the running the first policy on the first model. 
    
    '''
    jobs = itertools.product(model_structures, policies, scenarios)
    
    for i, job in enumerate(jobs):
        msi, policy, scenario = job
        name = '{} {} {}'.format(msi.name, policy.name, i)
        experiment = Experiment(name, msi.name, policy, scenario, i)
        yield experiment


def parameters_to_csv(parameters, file_name):
    '''Helper function for writing a collection of parameters to a csv file
    
    Parameters
    ----------
    parameters : collection of Parameter instances
    file_name :  str
    
    
    The function iterates over the collection and turns these into a data
    frame prior to storing them. The resulting csv can be loaded using the 
    create_parameters function. Note that currently we don't store resolution
    and default attributes. 
    
    '''
    
    params = {}
    
    for i, param in enumerate(parameters):
        
        if isinstance(param, CategoricalParameter):
            values = param.resolution
        else:
            values = param.lower_bound, param.upper_bound
        
        dict_repr = {j:value for j, value in enumerate(values)}
        dict_repr['name'] = param.name
        
        params[i] = dict_repr
        
    params = pandas.DataFrame.from_dict(params, orient='index')
    
    # for readability it is nice if name is the first column, so let's
    # ensure this
    cols = params.columns.tolist()
    cols.insert(0, cols.pop(cols.index('name')))
    params = params.reindex(columns=cols)

    # we can now safely write the dataframe to a csv
    pandas.DataFrame.to_csv(params, file_name, index=False)
    

def create_parameters(uncertainties, **kwargs):
    '''Helper function for creating many Parameters based on a DataFrame
    or csv file
    
    Parameters
    ----------
    uncertainties : str, DataFrame
    **kwargs : dict, arguments to pass to pandas.read_csv
    
    Returns
    -------
    list of Parameter instances
    
    
    This helper function creates uncertainties. It assumes that the 
    DataFrame or csv file has a column titled 'name', optionally a type column
    {int, real, cat}, can be included as well. the remainder of the columns
    are handled as values for the parameters. If type is not specified,
    the function will try to infer type from the values. 
    
    Note that this function does not support the resolution and default kwargs 
    on parameters. 
    
    An example of a csv:
    
    NAME,TYPE,,,
    a_real,real,0,1.1,
    an_int,int,1,9,
    a_categorical,cat,a,b,c
    
    this CSV file would result in 
    
    [RealParameter('a_real', 0, 1.1, resolution=[], default=None), 
     IntegerParameter('an_int', 1, 9, resolution=[], default=None), 
     CategoricalParameter('a_categorical', ['a', 'b', 'c'], default=None)]
    
    '''
    
    if isinstance(uncertainties, six.string_types):
        uncertainties = pandas.read_csv(uncertainties, **kwargs)
    elif not isinstance(uncertainties, pandas.DataFrame):
        uncertainties = pandas.DataFrame.from_dict(uncertainties)
    else:
        uncertainties = uncertainties.copy()
    
    parameter_map = {'int': IntegerParameter,
                     'real': RealParameter,
                     'cat': CategoricalParameter}

    # check if names column is there
    if ('NAME' not in uncertainties) and ('name' not in uncertainties):
        raise IndexError('name column missing')
    elif ('NAME' in uncertainties.columns):
        names = uncertainties.ix[:, 'NAME']
        uncertainties.drop(['NAME'], axis=1, inplace=True)
    else:
        names = uncertainties.ix[:, 'name']
        uncertainties.drop(['name'], axis=1, inplace=True)
    
    # check if type column is there
    infer_type = False
    if ('TYPE' not in uncertainties) and ('type' not in uncertainties):
        infer_type = True
    elif ('TYPE' in uncertainties):
        types = uncertainties.ix[:, 'TYPE']
        uncertainties.drop(['TYPE'], axis=1, inplace=True)
    else:
        types = uncertainties.ix[:, 'type']
        uncertainties.drop(['type'], axis=1, inplace=True)
    
    uncs = []
    for i, row in uncertainties.iterrows():
        name = names[i]
        values = row.values[row.notnull().values]
        type = None  # @ReservedAssignment
        
        if infer_type:
            if len(values) != 2:
                type = 'cat' # @ReservedAssignment
            else:
                l, u = values
                
                if isinstance(l, numbers.Integral) and isinstance(u, numbers.Integral):
                    type = 'int' # @ReservedAssignment
                else:
                    type = 'real' # @ReservedAssignment
            
        else:
            type = types[i] # @ReservedAssignment
            
            if (type != 'cat') and (len(values) != 2):
                raise ValueError('too many values specified for {}, is {}, should be 2'.format(name, values.shape[0]))
            
        if type=='cat':
            uncs.append(parameter_map[type](name, values))
        else:
            uncs.append(parameter_map[type](name, *values))
    return uncs




'''
This module specifies a generic ModelStructureInterface for controlling
NetLogo models. 
'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)
from ema_workbench.em_framework.model import Replicator, SingleReplication

try:
    import jpype
except ImportError:
    jpype = None
import os

import numpy as np
import six

from ..em_framework.model import FileModel
from ..util.ema_logging import method_logger
from ..util import warning, debug 

                         
from . import pyNetLogo

# Created on 15 mrt. 2013
# 
# .. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

__all__ = ['NetLogoModel']

class BaseNetLogoModel(FileModel):
    '''Base class for interfacing with netlogo models. This class
    extends :class:`em_framework.ModelStructureInterface`.
    
    Attributes
    ----------
    model_file : str
                 a relative path from the working directory to the model
    run_length : int
                 number of ticks
    command_format : str
                     default format for set operations in logo
    working_directory : str
    name : str
    
    '''
    command_format = "set {0} {1}"

    def __init__(self, name, wd=None, model_file=None):
        """
        init of class
        
        Parameters
        ----------
        working_directory : str
                            working_directory for the model. 
        name : str
               name of the modelInterface. The name should contain only
               alpha-numerical characters.
               
        Raises
        ------
        EMAError if name contains non alpha-numerical characters
        
        Note
        ----
        Anything that is relative to `self.working_directory`should be 
        specified in `model_init` and not in `src`. Otherwise, the code 
        will not work when running it in parallel. The reason for this is that 
        the working directory is being updated by parallelEMA to the worker's 
        separate working directory prior to calling `model_init`.
        
        """
        super(BaseNetLogoModel, self).__init__(name, wd=wd, model_file=model_file)
    
        self.run_length = None
       
    @method_logger
    def model_init(self, policy):
        '''
        Method called to initialize the model.
        
        Parameters
        ----------
        policy : dict
                 policy to be run.
        kwargs : dict
                 keyword arguments to be used by model_intit. This
                 gives users to the ability to pass any additional 
                 arguments. 
        
        '''
        super(BaseNetLogoModel, self).model_init(policy)
        debug("trying to start NetLogo")
        self.netlogo = pyNetLogo.NetLogoLink()
        debug("netlogo started")
        path = os.path.join(self.working_directory, self.model_file)
        self.netlogo.load_model(path)
        debug("model opened")
        
    @method_logger
    def run_experiment(self, experiment):
        """
        Method for running an instantiated model structure. 
        
        Parameters
        ----------
        experiment : dict like

        
        Raises
        ------
        jpype.JavaException if there is any exception thrown by the netlogo 
        model
        
        
        """
#         super(BaseNetLogoModel, self).run_model(scenario, policy)
        
        for key, value in experiment.items():
            try:
                self.netlogo.command(self.command_format.format(key, value))
            except jpype.JavaException as e:
                warning('variable {} throws exception: {}'.format(key,
                                                                  str(e)))
            
        debug("model parameters set successfully")
          
        # finish setup and invoke run
        self.netlogo.command("setup")
        
        # TODO:: it is possible to take advantage of of fact
        # that not all outcomes are time series
        # In that case, we need not embed the get command in the go
        # routine, but can do them at the end
        commands = []
        fns = {}
        for variable in self.outcome_variables:
            fn = r'{0}{3}{1}{2}'.format(self.working_directory,
                           variable,
                           ".txt",
                           os.sep)
            fns[variable] = fn
            fn = '"{}"'.format(fn)
            fn = fn.replace(os.sep, '/')
            
            if self.netlogo.report('is-agentset? {}'.format(variable)):
                # if name is name of an agentset, we
                # assume that we should count the total number of agents
                nc = r'{2} {0} {3} {4} {1}'.format(fn,
                                                   variable,
                                                   "file-open",
                                                   'file-write',
                                                   'count')
            else:
                # it is not an agentset, so assume that it is 
                # a reporter / global variable
                
                nc = r'{2} {0} {3} {1}'.format(fn,
                                               variable,
                                               "file-open",
                                               'file-write')
            commands.append(nc)
                

        c_start = "repeat {} [".format(self.run_length)
        c_close = "go ]"
        c_middle = " ".join(commands)
#         c_end = " ".join(end_commands)
        command = " ".join((c_start, c_middle, c_close))
        debug(command)
        self.netlogo.command(command)
        
        # after the last go, we have not done a write for the outcomes
        # so we do that now
        self.netlogo.command(c_middle)
        
        # we also need to save the non time series outcomes
#         self.netlogo.command(c_end)
        
        self.netlogo.command("file-close-all")
        return self._handle_outcomes(fns)

    def retrieve_output(self):
        """
        Method for retrieving output after a model run.
        
        Returns
        -------
        dict with the results of a model run. 
        
        """
        return self.output
    
    def cleanup(self):
        '''
        This model is called after finishing all the experiments, but 
        just prior to returning the results. This method gives a hook for
        doing any cleanup, such as closing applications. 
        
        In case of running in parallel, this method is called during 
        the cleanup of the pool, just prior to removing the temporary 
        directories. 
        
        '''
        self.netlogo.kill_workspace()
        jpype.shutdownJVM()

    def _handle_outcomes(self, fns):
        '''helper function for parsing outcomes'''
        
        results = {}
        for key, value in fns.items():
            with open(value) as fh:
                result = fh.readline()
                result = result.strip()
                result = result.split()
                result = [float(entry) for entry in result]
                results[key] = np.asarray(result)
            os.remove(value)        
        
        temp_output = {}
        for outcome in self.outcomes:
            varname = outcome.variable_name
            if len(varname)==1:
                varname = varname[0]
                temp_output[outcome.name] = results[varname]
            else:
                temp_output[outcome.name] = [results[var] for var in varname]
        return temp_output
         
class NetLogoModel(Replicator, BaseNetLogoModel):
    pass

class SingleReplicationNetLogoModel(SingleReplication, BaseNetLogoModel):
    pass


'''

This module provides an abstract base class for a callback and a default
implementation.

If you want to store the data in a way that is different from the 
functionality provided by the default callback, you can write your own 
extension of callback. For example, you can easily implement a callback
that stores the data in e.g. a NoSQL file.  

The only method to implement is the __call__ magic method. To use logging of
progress, always call super. 

'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)

import abc
from threading import Lock

import numpy as np

from ..util import ema_logging, ema_exceptions
from .parameters import CategoricalParameter, IntegerParameter

#
# Created on 22 Jan 2013
#
# .. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>
#

__all__ = ['AbstractCallback',
           'DefaultCallback']

class AbstractCallback(object):
    '''
    Abstract base class from which different call back classes can be derived.
    Callback is responsible for storing the results of the runs.

    Parameters
    ----------
    uncs : list
            a list of the parameters over which the experiments 
            are being run.
    outcomes : list
               a list of outcomes
    nr_experiments : int
                     the total number of experiments to be executed
    reporting_interval : int, optional 
                         the interval at which to provide progress information 
                         via logging.

    Attributes
    ----------
    i : int
        a counter that keeps track of how many experiments have been saved
    reporting_interval : int
                         the frequency at which to log progress

    '''
    __metaclass__ = abc.ABCMeta
    
    i = 0
    reporting_interval = 100
    
    def __init__(self, 
                 uncertainties, 
                 outcomes,
                 levers,
                 nr_experiments,
                 reporting_interval=None):
        
        if reporting_interval is None:
            reporting_interval = max(1, int(round(nr_experiments / 10))) 
        
        self.reporting_interval = reporting_interval
            
    @abc.abstractmethod
    def __call__(self, experiment, result):
        '''
        Method responsible for storing results. The implementation in this
        class only keeps track of how many runs have been completed and 
        logging this. Any extension of AbstractCallback needs to implement
        this method. If one want to use the logging provided here, call it via
        super.
        
        Parameters
        ----------
        experiment: Experiment instance
        result: dict
                the result dict
        
        '''
        #
        # TODO:: replace with optional tqdm based progress bar
        # http://thelivingpearl.com/2012/12/31/creating-progress-bars-with-python/ 
        # idea: have a bar that ships with workbench, which is used as a 
        # fallback if tqdm is not available
        self.i+=1
        ema_logging.debug(str(self.i)+" cases completed")
        
        if self.i % self.reporting_interval == 0:
            ema_logging.info(str(self.i)+" cases completed")

    @abc.abstractmethod
    def get_results(self):
        """
        method for retrieving the results. Called after all experiments have 
        been completed. Any extension of AbstractCallback needs to implement
        this method.
        """

        
class DefaultCallback(AbstractCallback):
    """ 
    default callback system
    callback can be used in performExperiments as a means for specifying 
    the way in which the results should be handled. If no callback is 
    specified, this default implementation is used. This one can be 
    overwritten or replaced with a callback of your own design. For 
    example if you prefer to store the result in a database or write 
    them to a text file
    """
    
    i = 0
    cases = None 
    results = {}
    
    shape_error_msg = "can only save up to 2d arrays, this array is {}d"
    
    def __init__(self, 
                 uncs, 
                 levers,
                 outcomes, 
                 nr_experiments, 
                 reporting_interval=100):
        '''
        
        
        Parameters
        ----------
        uncs : list
                a list of the parameters over which the experiments 
                are being run.
        outcomes : list
                   a list of outcomes
        nr_experiments : int
                         the total number of experiments to be executed
        reporting_interval : int 
                             the interval at which to provide
                             progress information via logging.
        
        '''
        super(DefaultCallback, self).__init__(uncs, 
                                              levers,
                                              outcomes, 
                                              nr_experiments, 
                                              reporting_interval)
        self.i = 0
        self.cases = None
        self.results = {}
        self.lock = Lock()
        
        self.outcomes = [outcome.name for outcome in outcomes]

        #determine data types of parameters
        self.dtypes = []
        self.parameters = []
        
        for parameter in uncs + levers:
            name = parameter.name
            self.parameters.append(name)
            dataType = float
            
            if isinstance(parameter, CategoricalParameter):
                dataType = object
            elif isinstance(parameter, IntegerParameter):
                dataType = int
            self.dtypes.append((str(name), dataType))
        self.dtypes.append((str('scenario_id'), object))
        self.dtypes.append((str('policy'), object))
        self.dtypes.append((str('model'), object))
        
        
        self.cases = np.empty((nr_experiments,), dtype=self.dtypes)
        self.cases[:] = np.NAN
        self.nr_experiments = nr_experiments

    def _store_case(self, experiment):
        scenario = experiment.scenario
        policy = experiment.policy
        
        case = []
        for parameter in self.parameters:
            try:
                value = scenario[parameter]
            except KeyError:
                try:
                    value = policy[parameter]
                except KeyError:
                    value = np.nan
            finally:
                case.append(value)
        
        case.append(scenario.name)
        case.append(policy.name)
        case.append(experiment.model_name)
        case = tuple(case)
        self.cases[experiment.experiment_id] = case
            
    def _store_result(self, case_id, result):
        for outcome in self.outcomes:
            ema_logging.debug("storing {}".format(outcome))
            
            try:
                outcome_res = result[outcome]
            except KeyError:
                ema_logging.debug("%s not specified as outcome in msi" % outcome)
            else:
                try:
                    self.results[outcome][case_id, ] = outcome_res
                except KeyError: 
                    shape = np.asarray(outcome_res).shape
                    
                    if len(shape)>2:
                        raise ema_exceptions.EMAError(self.shape_error_msg.format(len(shape)))
                    
                    shape = list(shape)
                    shape.insert(0, self.nr_experiments)
                    
                    self.results[outcome] = np.empty(shape)
                    self.results[outcome][:] = np.NAN
                    self.results[outcome][case_id, ] = outcome_res
    
    def __call__(self, experiment, result ):
        '''
        Method responsible for storing results. This method calls 
        :meth:`super` first, thus utilizing the logging provided there
        
        
        
        Parameters
        ----------
        experiment: Experiment instance
        result: dict
                the result dict
        
        '''
        super(DefaultCallback, self).__call__(experiment, result)

        self.lock.acquire()
                           
        #store the case
        self._store_case(experiment)
        
        #store results
        self._store_result(experiment.experiment_id, result)
        
        self.lock.release()
        
    def get_results(self):
        return self.cases, self.results


'''
Created on Jul 8, 2014

@author: jhkwakkel@tudelft.net
'''
import matplotlib.pyplot as plt

from ema_workbench import ema_logging, load_results

from ema_workbench.analysis.plotting import envelopes 
from ema_workbench.analysis.plotting_util import KDE

ema_logging.log_to_stderr(ema_logging.INFO)

file_name = r'./data/1000 flu cases.tar.gz'
results = load_results(file_name)

# the plotting functions return the figure and a dict of axes
fig, axes = envelopes(results, group_by='policy', density=KDE, fill=True)

# we can access each of the axes and make changes
for key, value in axes.iteritems():
    # the key is the name of the outcome for the normal plot
    # and the name plus '_density' for the endstate distribution
    if key.endswith('_density'):
        value.set_xscale('log')

plt.show()


'''
Created on Dec 19, 2011

@author: gyucel
'''

import numpy as np


from ema_workbench.util import info

def ssedist(d1,d2):
    d = ((d1-d2)**2).sum()
#    print d
    return d

def distance_sse(data):
    
    '''
    The SSE (sum of squared-errors) distance between two data series is equal to the sum of squared-errors between corresponding data points of these two data series.
    Let the data series be of length N; Then SSE distance between ds1 and ds2 equals to the sum of the square of error terms from 1 to N, 
    where error_term(i) equals to ds1(i)-ds2(i) 
    
    Since SSE calculation is based on pairwise comparison of individual data points, the data series should be of equal length.
    
    SSE distance equals to the square of Euclidian distance, which is a commonly used distance metric in time series comparisons.
    '''
    
    runLogs = []
    #Generates the feature vectors for all the time series that are contained in numpy array data
    info("calculating distances")
    dRow = np.zeros(shape=(np.sum(np.arange(data.shape[0])), ))
    index = -1
    for i in range(data.shape[0]):
            
        # For each run, a log is created
        # Log includes a description dictionary that has key information 
        # for post-clustering analysis, and the data series itself. These 
        # logs are stored in a global array named runLogs
        behaviorDesc = {}
        behaviorDesc['Index'] = str(i)
        
        behavior = data[i]
        localLog = (behaviorDesc, behavior)
        runLogs.append(localLog)
    
        for j in range(i+1, data.shape[0]):
            index += 1
            distance = ssedist(data[i],data[j]) 
            dRow[index] = distance
    return dRow, runLogs

if __name__ == '__main__':
    tester = np.array([(12,4),(2,2)])



'''

Module offers support for performing basic regional sensitivity analysis. The
module can be used to perform regional sensitivity analysis on all 
uncertainties specified in the experiment array, as well as the ability to 
zoom in on any given uncertainty in more detail. 


'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)

import operator

import matplotlib.pyplot as plt
import numpy as np
import numpy.lib.recfunctions as rf
import seaborn as sns

# Created on Aug 18, 2015
# 
# .. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

__all__ = ['plot_cdf',
           'plot_cdfs']

cp = sns.color_palette()

def build_legend(x,y):
    '''helper function for building a legend
    
    Parameters
    ----------
    x : ndarray
    y : ndarray
    
    '''
    proxies = []
    labels = []
    for i in range(np.max(y)+1):
        proxy = plt.Line2D([0,1], [0,1], color=cp[i+1])
        proxies.append(proxy) 
        labels.append('{} (N={})'.format(i,x[y==i].shape[0]))
    proxies.append(plt.Line2D([0,1], [0,1], lw=1,color='darkgrey'))
    labels.append('unconditioned')
    return proxies, labels


def plot_discrete_cdf(ax, unc, x, y, xticklabels_on,
                     ccdf):
    '''plot a discrete cdf on ax for data,
    grouping data by logical index.
    
    Parameters
    ----------
    ax : matplotlib axes
    unc : str
    x  : ndarray
    y : ndarray
    xticklabels_on : bool
    ccdf : bool
    
    '''
    cats = sorted(set(x))
    n_cat = len(cats)
    for i in range(np.max(y)+1):
        data_i = x[y==i]

        freqs = []
        for cat in cats:
            freq = data_i[data_i==cat].shape[0]/data_i.shape[0]
            freqs.append((cat, freq))
            
        freqs.sort(key=operator.itemgetter(1))
        cats = map(operator.itemgetter(0), freqs)
        freqs = map(operator.itemgetter(1), freqs)
        
        cum_freq = 0
        for j, freq in enumerate(freqs):
            cum_freq += freq
            
            freq = cum_freq
            
            if ccdf:
                freq = 1 - cum_freq

                
            x_plot = [j*1, j*1+1]
            y_plot = [freq,]*2
            
            ax.plot(x_plot, y_plot ,c=cp[i+1], label=i==1)
            ax.scatter(x_plot[0], y_plot[0], edgecolors=cp[i+1], facecolors=cp[i+1],
                      linewidths=1, zorder=2)
            ax.scatter(x_plot[1], y_plot[0], edgecolors=cp[i+1], facecolors='white',
                      linewidths=1, zorder=2)
            
            # misnomer
            cum_freq_un = (j+1)/n_cat
            if ccdf:
                cum_freq_un = (len(freqs)-j-1)/n_cat
            
            ax.plot(x_plot, [cum_freq_un,]*2, lw=1, c='darkgrey',
                   zorder=1, label='x==y')
            ax.scatter(x_plot[0], cum_freq_un, edgecolors='darkgrey', 
                       facecolors='darkgrey', linewidths=1,
                       zorder=1)
            ax.scatter(x_plot[1], cum_freq_un, edgecolors='darkgrey',
                       facecolors='white', linewidths=1,
                      zorder=1)
            

    ax.set_xticklabels([])
    if xticklabels_on:
        for i,cat in enumerate(cats):
            ax.text(i*1+0.5,-0.1, cat, ha='center', rotation=45)

    ax.set_ylim(ymin=-0.01, ymax=1.01)

    xmin=-0.02*n_cat
    xmax=n_cat+0.02*n_cat
    ax.set_xticks(np.linspace(xmin, xmax, 4))
    ax.set_xlim(xmin=xmin, xmax=xmax)

    
def plot_continuous_cdf(ax, unc, x, y, xticklabels_on,
                       ccdf):
    '''plot a continuous cdf on ax for data,grouping data by the groups
    specified in y.
    
    Parameters
    ----------
    ax : matplotlib axes 
    unc : str
    x  : ndarray
    y : ndarray
    xticklabels_on : bool
    ccdf : bool
    
    '''
    
    for i in range(np.max(y)+1):
        data_i = x[y==i]
        sorted_data = np.sort(data_i)
        yvals = np.arange(len(sorted_data))/float(len(sorted_data))
        if ccdf:
            yvals = 1 - yvals
        ax.plot(sorted_data,yvals, color=cp[i+1], label='{}'.format(i))
    
    x0 = min(x)
    x1 = max(x)
    
    sorted_data = np.sort(x)
    yvals = np.arange(len(x))/float(len(x))
    if ccdf:
        yvals = 1 - yvals
    
    ax.plot(sorted_data, yvals, c='darkgrey', lw=1)         
    
    ax.set_xlim(xmin=x0, xmax=x1)
    xticklocs = np.linspace(x0, x1, 4)
    ax.set_xticks(xticklocs)
    if xticklabels_on:
        ax.set_xticklabels(['{:.2g}'.format(entry) for entry in xticklocs])
    else:
        ax.set_xticklabels([])

        
def plot_cdf(ax, unc, x, y, discrete=False,
            legend=False, xticklabels_on=False,
            yticklabels_on=False, ccdf=False):
    '''plot cdf for x conditional on y
    
    Parameters
    ----------
    ax : Axes instance
         axes on which to plot the cdf
    unc : str
          the name of the uncertainty
    x : ndarray of shape (1,)
        the data to plot
    y : ndarray(1,)
        the categorization for the data
    discrete : bool, optional
               if true, plot a discrete cdf. Default is false.
    legend : bool, optional
    xticklabels_on : bool, optional
    ccdf : bool, optional
           if true, plot a complementary cdf instead of a normal cdf.
    
    '''

    if discrete:
        plot_discrete_cdf(ax, unc, x, y, xticklabels_on,
                         ccdf)
    else:
        plot_continuous_cdf(ax, unc, x, y, xticklabels_on,
                           ccdf)

    if legend:
        proxies, labels = build_legend(x,y)
        ax.legend(proxies, labels, loc='best')

    yticklocs = np.linspace(0,1,4)
    ax.set_yticks(yticklocs)
    
    if yticklabels_on:
        ax.set_yticklabels(['$0$', '$\\frac{1}{3}$', '$\\frac{2}{3}$', '$1$'])
    else:
        ax.set_yticklabels([])
        
    if xticklabels_on:
        ax.set_xlabel(str(unc))
    else:
        x0, x1 = ax.get_xlim()
        ax.text(x0+0.01*x1, 1, str(unc), va='top', ha='left')
        

def plot_cdfs(x, y, ccdf=False):
    '''plot cumulative density functions for each column in x, based on the 
    classification specified in y.
    
    Parameters
    ----------
    x : recarray
        the experiments to use in the cdfs
    y : ndaray 
        the categorization for the data
    ccdf : bool, optional
           if true, plot a complementary cdf instead of a normal cdf.
    
    '''
    x = rf.drop_fields(x, "scenario_id", asrecarray=True)
    uncs = rf.get_names(x.dtype)
    cp = sns.color_palette()
    
    n_col = 4
    n_row = len(uncs)//n_col +1
    size = 3 
    aspect = 1
    figsize = n_col * size * aspect, n_row * size
    fig, axes = plt.subplots(n_row, n_col,
                             figsize=figsize,
                             squeeze=False)

    for i, unc in enumerate(uncs):
        discrete = False
        
        i_col = i % n_col
        i_row = i // n_col
        ax = axes[i_row, i_col]
        
        data = x[unc]
        if x.dtype[unc] == np.dtype('O'):
            discrete = True
        plot_cdf(ax, unc, data, y, discrete, ccdf=ccdf)
    
    # last row might contain empty axis, 
    # let's make them disappear
    i_row = len(uncs) // n_col
    i_col = len(uncs) % n_col
    for i_col in range(i_col, n_col):
        ax = axes[i_row, i_col]
        ax.set_xticklabels([])
        ax.set_xticks([])
        ax.set_yticklabels([])
        ax.set_yticks([])
        
        sns.despine(ax=ax, top=True, right=True, left=True, bottom=True)
    
    proxies, labels = build_legend(x, y)
    
    fig.legend(proxies, labels, "upper center")

    return fig


'''
Created on Dec 19, 2011

@author: gyucel
'''
import numpy as np

from ema_workbench.util import info

def msedist(d1,d2):
    sse = ((d1-d2)**2).sum()
    mse = np.average(sse)
    return mse

def distance_mse(data):
    '''
    The MSE (mean squared-error) distance is equal to the SSE distance divided by the number of data points in data series.
    
    The SSE distance between two data series is equal to the sum of squared-errors between corresponding data points of these two data series.
    Let the data series be of length N; Then SSE distance between ds1 and ds2 equals to the sum of the square of error terms from 1 to N, 
    where error_term(i) equals to ds1(i)-ds2(i) 
    
    Given that SSE is calculated as given above, MSE equals SSE divided by N.
    
    As SSE distance, the MSE distance only works with data series of equal length.
    '''
    
    runLogs = []
    #Generates the feature vectors for all the time series that are contained in numpy array data
    info("calculating distances")
    dRow = np.zeros(shape=(np.sum(np.arange(data.shape[0])), ))
    index = -1
    for i in range(data.shape[0]):
            
        # For each run, a log is created
        # Log includes a description dictionary that has key information 
        # for post-clustering analysis, and the data series itself. These 
        # logs are stored in a global array named runLogs
        behaviorDesc = {}
        behaviorDesc['Index'] = str(i)
        
        behavior = data[i]
        localLog = (behaviorDesc, behavior)
        runLogs.append(localLog)
    
        for j in range(i+1, data.shape[0]):
            index += 1
            distance = msedist(data[i],data[j]) 
            dRow[index] = distance
    return dRow, runLogs

if __name__ == '__main__':
    tester = np.array([(12,4),(2,2)])


'''

This module provides various convenience functions and classes.

'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)

try:
    import configparser
except ImportError:
    import ConfigParser as configparser

# import cPickle
from io import BytesIO, StringIO
import math
import os
import sys
import tarfile

from matplotlib.mlab import rec2csv
import numpy as np
from numpy.lib import recfunctions

import pandas as pd
from pandas.io.parsers import read_csv

from .ema_logging import info, debug, warning
from .ema_exceptions import EMAError

PY3 = sys.version_info[0] == 3
if PY3:
    WriterFile =  StringIO
else:
    WriterFile =  BytesIO

# Created on 13 jan. 2011
# 
# .. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

__all__ = ['load_results',
           'save_results',
           'experiments_to_cases',
           'merge_results'
           ]


def load_results(file_name):
    '''
    load the specified bz2 file. the file is assumed to be saves
    using save_results.
    
    Parameters
    ----------    
    file_name : str
                the path to the file
                
    Raises
    ------
    IOError if file not found

    '''
    file_name = os.path.abspath(file_name)
    outcomes = {}
    with tarfile.open(file_name, 'r:gz', encoding="UTF8") as z:
        # load x
        experiments = z.extractfile('experiments.csv')
        if not (hasattr(experiments, 'read')):
            raise EMAError(repr(experiments))
        
        df = pd.read_csv(experiments)
        experiments = df.to_records(index=False)
#         experiments = recfunctions.drop_fields(experiments, ['index'])

        # load experiment metadata
        metadata = z.extractfile('experiments metadata.csv').readlines()
        
        metadata_temp = []
        for entry in metadata:
            entry = entry.decode('UTF-8')
            entry = entry.strip()
            entry = entry.split(",")
            entry = [str(item) for item in entry]
            entry = tuple(entry)
            metadata_temp.append(entry)
        metadata = metadata_temp    
        
        metadata = np.dtype(metadata)

        # cast experiments to dtype and name specified in metadata        
        temp_experiments = np.zeros((experiments.shape[0],), dtype=metadata)
        for i, entry in enumerate(experiments.dtype.descr):
            dtype = metadata[i]
            name = metadata.descr[i][0]
            temp_experiments[name][:] = experiments[entry[0]].astype(dtype)
        experiments = temp_experiments
        
        # load outcome metadata
        metadata = z.extractfile('outcomes metadata.csv').readlines()
        metadata = [entry.decode('UTF-8') for entry in metadata]
        metadata = [entry.strip() for entry in metadata]
        metadata = [tuple(entry.split(",")) for entry in metadata]
        metadata = {entry[0]: entry[1:] for entry in metadata}

        # load outcomes
        for outcome, shape in metadata.items():
            shape = list(shape)
            shape[0] = shape[0][1:]
            shape[-1] = shape[-1][0:-1]
            
            temp_shape = []
            for entry in shape:
                if entry:
                    try:
                        temp_shape.append(int(entry))
                    except ValueError:
                        try:
                            temp_shape.append(int(long(entry))) #@UndefinedVariable
                        except NameError: # we are on python3
                            temp_shape.append(int(entry[0:-1]))
            shape = tuple(temp_shape)
            
            if len(shape)>2:
                nr_files = shape[-1]
                
                data = np.empty(shape)
                for i in range(nr_files):
                    values = z.extractfile("{}_{}.csv".format(outcome, i))
                    values = read_csv(values, index_col=False, header=None).values
                    data[:,:,i] = values

            else:
                data = z.extractfile("{}.csv".format(outcome))
                data = read_csv(data, index_col=False, header=None).values
                data = np.reshape(data, shape)
                
            outcomes[outcome] = data
            
    info("results loaded succesfully from {}".format(file_name))
    return experiments, outcomes


def save_results(results, file_name):
    '''
    save the results to the specified tar.gz file. The results are stored as 
    csv files. There is an x.csv, and a csv for each outcome. In 
    addition, there is a metadata csv which contains the datatype information
    for each of the columns in the x array.

    Parameters
    ----------    
    results : tuple
              the return of run_experiments
    file_name : str
                the path of the file
    
    Raises
    ------
    IOError if file not found

    '''
    file_name = os.path.abspath(file_name)

    def add_file(tararchive, string_to_add, filename):
        tarinfo = tarfile.TarInfo(filename)
        tarinfo.size = len(string_to_add)
        
        fh = BytesIO(string_to_add.encode('UTF-8'))
        
        z.addfile(tarinfo, fh)  
    
    def save_numpy_array(fh, data):
        data = pd.DataFrame(data)
        data.to_csv(fh, header=False, index=False, encoding='UTF-8')
        
    experiments, outcomes = results
    with tarfile.open(file_name, 'w:gz') as z:
        # write the x to the zipfile
        experiments_file = WriterFile()
        rec2csv(experiments, experiments_file, withheader=True)
        add_file(z, experiments_file.getvalue(), 'experiments.csv')
        
        # write experiment metadata
        dtype = experiments.dtype.descr
        dtype = ["{},{}".format(*entry) for entry in dtype]
        dtype = "\n".join(dtype)
        add_file(z, dtype, 'experiments metadata.csv')
        
        # write outcome metadata
        outcome_names = outcomes.keys()
        outcome_meta = ["{},{}".format(outcome, outcomes[outcome].shape) 
                        for outcome in outcome_names]
        outcome_meta = "\n".join(outcome_meta)
        add_file(z, outcome_meta, "outcomes metadata.csv")
        
        # outcomes
        for key, value in outcomes.items():
            fh = WriterFile()
            
            nr_dim = len(value.shape)
            if nr_dim==3:
                for i in range(value.shape[2]):
                    data = value[:,:,i]
                    save_numpy_array(fh, data)
                    fh = fh.getvalue()
                    fn = '{}_{}.csv'.format(key, i)
                    add_file(z, fh, fn)
                    fh = WriterFile()
            else:
                save_numpy_array(fh, value)
                fh = fh.getvalue()
                add_file(z, fh, '{}.csv'.format(key))
  
    info("results saved successfully to {}".format(file_name))
    

def experiments_to_cases(experiments):
    '''
    
    This function transform a structured x array into a list
    of case dicts. This can then for example be used as an argument for 
    running :meth:`~em_framework.model_ensemble.ModelEnsemble.perform_experiments`.
    
    Parameters
    ----------    
    experiments : numpy structured array
                  a structured array containing experiments
    
    Returns
    -------
    a list of case dicts.
    
    '''
    #get the names of the uncertainties
    uncertainties = [entry[0] for entry in experiments.dtype.descr]
    
    #remove policy and model, leaving only the case related uncertainties
    try:
        uncertainties.pop(uncertainties.index('policy'))
        uncertainties.pop(uncertainties.index('model'))
    except:
        pass
    
    #make list of of tuples of tuples
    cases = []
    cache = set()
    for i in range(experiments.shape[0]):
        case = {}
        case_tuple = []
        for uncertainty in uncertainties:
            entry =  experiments[uncertainty][i]
            case[uncertainty] = entry
            case_tuple.append(entry)
        
        case_tuple = tuple(case_tuple)
        if case_tuple not in cache:
            cases.append(case)
            cache.add((case_tuple))
    
    return cases


def merge_results(results1, results2, downsample=None):
    '''
    convenience function for merging the return from 
    :meth:`~modelEnsemble.ModelEnsemble.perform_experiments`.
    
    The function merges results2 with results1. For the experiments,
    it generates an empty array equal to the size of the sum of the 
    experiments. As dtype is uses the dtype from the experiments in results1.
    The function assumes that the ordering of dtypes and names is identical in
    both results.  
    
    A typical use case for this function is in combination with 
    :func:`~util.experiments_to_cases`. Using :func:`~util.experiments_to_cases`
    one extracts the cases from a first set of experiments. One then
    performs these cases on a different model or policy, and then one wants to
    merge these new results with the old result for further analysis.  
    
    Parameters
    ----------
    results1 : tuple
               first results to be merged
    results2 : tuple
               second results to be merged
    downsample : int 
                 should be an integer, will be used in slicing the results
                 in order to avoid memory problems. 

    Returns
    -------
    the merged results
    
    
    '''

    #start of merging
    old_exp, old_res = results1
    new_exp, new_res = results2
    
    #merge x
    dtypes = old_exp.dtype
    
    merged_exp = np.empty((old_exp.shape[0]+new_exp.shape[0],),dtype= dtypes)
    merged_exp[0:old_exp.shape[0]] = old_exp
    merged_exp[old_exp.shape[0]::] = new_exp
    
    #only merge the results that are in both
    keys = old_res.keys()
    [keys.append(key) for key in new_res.keys()]
    keys = set(keys)
    info("intersection of keys: %s" % keys)
    
    #merging results
    merged_res = {}
    for key in keys:
        info("merge "+key)
        
        old_value = old_res.get(key)
        new_value = new_res.get(key)
        
        i = old_value.shape[0]+new_value.shape[0]
        j = old_value.shape[1]
        slice_value = 1
        if downsample:
            j = int(math.ceil(j/downsample))
            slice_value = downsample
            
        merged_value = np.empty((i,j))
        debug("merged shape: %s" % merged_value.shape)
        
        merged_value[0:old_value.shape[0], :] = old_value[:, ::slice_value]
        merged_value[old_value.shape[0]::, :] = new_value[:, ::slice_value]

        merged_res[key] = merged_value
    
    mr = (merged_exp, merged_res)
    return mr  


def get_ema_project_home_dir():
    try:
        config_file_name = "expworkbench.cfg"
        directory = os.path.dirname(__file__)
        fn = os.path.join(directory, config_file_name)

        config = configparser.ConfigParser()
        parsed = config.read(fn)

        if parsed:
            info('config loaded from {}'.format(parsed[0]))
        else:
            info('no config file found')


        home_dir = config.get('ema_project_home', 'home_dir')
        return home_dir
    except:
        return os.getcwd()



'''
Created on 3 Jan. 2011

This file illustrated the use the EMA classes for a contrived vensim
example


.. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>
                chamarat <c.hamarat (at) tudelft (dot) nl>
'''
from __future__ import (division, unicode_literals, absolute_import, 
                        print_function)

from ema_workbench import (TimeSeriesOutcome, perform_experiments,
                           RealParameter, ema_logging)

from ema_workbench.connectors.vensim import VensimModel

if __name__ == "__main__":
    #turn on logging
    ema_logging.log_to_stderr(ema_logging.INFO)
    
    #instantiate a model
    wd = r'./models/vensim example'
    vensimModel = VensimModel("simpleModel", wd=wd,
                              model_file=r'\model.vpm')
    vensimModel.uncertainties = [RealParameter("x11", 0, 2.5),
                                 RealParameter("x12", -2.5, 2.5)]
    
    vensimModel.outcomes = [TimeSeriesOutcome('a')]
    
    results = perform_experiments(vensimModel, 1000, parallel=True)
    


'''

Exceptions and warning used internally by the EMA workbench. In line with 
advice given in `PEP 8 <http://www.python.org/dev/peps/pep-0008/>`_.
'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)

# Created on 31 mei 2011
# 
# .. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

__all__ = ['EMAError',
           'EMAWarning',
           'CaseError',
           'EMAParallelError']


class EMAError(BaseException):
    '''
    Base EMA error
    '''
    def __init__(self, *args):
        self.args = args

    def __str__(self):
        if len(self.args) == 1:
            return str(self.args[0])
        else:
            return str(self.args)

    def __repr__(self):
        return "%s(*%s)" % (self.__class__.__name__, repr(self.args))
    
class EMAWarning(EMAError):
    '''
    base EMA warning class
    '''
    pass

class CaseError(EMAError):
    '''
    error to be used when a particular run creates an error. The character of 
    the error can be specified as the message, and the actual case that 
    gave rise to the error. 
    
    '''
    def __init__(self, message, case, policy=None):
        self.message = message
        self.case = case
        self.args = (message, case)
        try:
            self.policy = policy.name
        except AttributeError:
            self.policy = 'None'
    
    def __str__(self):
        keys = sorted(self.case.keys())
        
        
        c = ""
        for key in keys:
            value = self.case.get(key)
            c += key
            c += ":"
            c += str(value)
            c += ', '
        c+= 'policy:'+self.policy
        
        return self.message + ' case: {' + c + "}"

    def __repr__(self):
        return "%s case: %s " % (self.message, repr(self.case))        

class EMAParallelError(EMAError):
    '''
    parallel EMA error
    '''
    pass


'''

This module provides R style pairs plotting functionality.

'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)

import matplotlib.cm as cm
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np

from . import plotting_util
from .plotting_util import SCATTER, LINE
from .plotting_util import prepare_pairs_data, make_legend
from ..util import debug, info

# .. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

__all__ = ['pairs_scatter', 'pairs_lines', 'pairs_density']


def pairs_lines(results, 
                outcomes_to_show = [],
                group_by = None,
                grouping_specifiers = None,
                ylabels = {},
                legend=True,
                **kwargs):
    '''
    
    Generate a `R style pairs <http://www.stat.psu.edu/~dhunter/R/html/graphics/html/pairs.html>`_ 
    lines multiplot. It shows the behavior of two outcomes over time against
    each other. The origin is denoted with a circle and the end is denoted
    with a '+'. 
    
    Parameters
    ----------
    results : tuple
              return from perform_experiments.
    outcomes_to_show : list of str, optional
                       list of outcome of interest you want to plot.
    group_by : str, optional
               name of the column in the cases array to group results by. 
               Alternatively, `index` can be used to use indexing arrays as the 
               basis for grouping.
    grouping_specifiers : dict, optional
                          dict of categories to be used as a basis for grouping 
                          by. Grouping_specifiers is only meaningful if 
                          group_by is provided as well. In case of grouping by
                          index, the grouping  specifiers should be in a 
                          dictionary where the key denotes the name of the 
                          group. 
    ylabels : dict, optional
              ylabels is a dictionary with the outcome names as keys, the 
              specified values will be used as labels for the y axis. 
    legend : bool, optional
             if true, and group_by is given, show a legend.
    point_in_time : float, optional
                    the point in time at which the scatter is to be made. If 
                    None is provided (default), the end states are used. 
                    point_in_time should be a valid value on time
    
    Returns
    -------
    fig
        the figure instance
    dict
        key is tuple of names of outcomes, value is associated axes
        instance
    
    '''
    
    #unravel return from run_experiments   
    debug("making a pars lines plot")
    
    prepared_data = prepare_pairs_data(results, outcomes_to_show, group_by, 
                                       grouping_specifiers, None)
    outcomes, outcomes_to_show, grouping_labels = prepared_data
    
    grid = gridspec.GridSpec(len(outcomes_to_show), len(outcomes_to_show))                             
    grid.update(wspace = 0.1,
                hspace = 0.1)
    
    #the plotting
    figure = plt.figure()
    axes_dict = {}
  
    combis = [(field1, field2) for field1 in outcomes_to_show\
                               for field2 in outcomes_to_show]
    
    for field1, field2 in combis:
        i = list(outcomes_to_show).index(field1)
        j = list(outcomes_to_show).index(field2)
        ax = figure.add_subplot(grid[i,j])
        
        axes_dict[(field1, field2)] = ax

        if group_by:
            for x, entry in enumerate(grouping_labels):
                data1 = outcomes[entry][field1]
                data2 = outcomes[entry][field2]
                color = plotting_util.COLOR_LIST[x]
                if i==j: 
                    color = 'white'
                simple_pairs_lines(ax, data1, data2, color)
        else:
            data1 = outcomes[field1]
            data2 = outcomes[field2]
            color = 'b'
            if i==j: 
                color = 'white'
            simple_pairs_lines(ax, data1, data2, color)
        do_text_ticks_labels(ax, i, j, field1, field2, ylabels, 
                             outcomes_to_show)

    if group_by and legend:
        gs1 = grid[0,0]
        
        for ax in figure.axes:
            gs2 = ax._subplotspec
            if all((gs1._gridspec == gs2._gridspec,
                    gs1.num1 == gs2.num1,
                    gs1.num2 == gs2.num2)):
                break  
        
        make_legend(grouping_labels, ax, legend_type=LINE)

    return figure, axes_dict
 
 
def simple_pairs_lines(ax, y_data, x_data, color):    
    '''
    
    Helper function for generating a simple pairs lines plot

    Parameters
    ----------
    ax : axes
    data1 : ndarray
    data2 : ndarray
    color : str
    
    '''
             
    ax.plot(x_data.T, y_data.T, c=color)
    ax.scatter(x_data[:, 0], y_data[:, 0],
               edgecolor=color, facecolor=color,
               marker='o')
    ax.scatter(x_data[:, -1], y_data[:, -1],
               edgecolor=color, facecolor=color,
               marker='+')     


def pairs_density(results, 
                  outcomes_to_show = [],
                  group_by = None,
                  grouping_specifiers = None,
                  ylabels = {},
                  point_in_time=-1,
                  log=True,
                  gridsize=50,
                  colormap='coolwarm',
                  filter_scalar=True): 
    '''
    
    Generate a `R style pairs <http://www.stat.psu.edu/~dhunter/R/html/graphics/html/pairs.html>`_ 
    hexbin density multiplot. In case of time-series data, the end states are 
    used.
    
    hexbin makes hexagonal binning plot of x versus y, where x, y are 1-D 
    sequences of the same length, N. If C is None (the default), this is a 
    histogram of the number of occurences of the observations at (x[i],y[i]).
    For further detail see `matplotlib on hexbin <http://matplotlib.sourceforge.net/api/pyplot_api.html#matplotlib.pyplot.hexbin>`_

    Parameters
    ----------
    results : tuple
              return from perform_experiments.
    outcomes_to_show : list of str, optional
                       list of outcome of interest you want to plot.
    group_by : str, optional
               name of the column in the cases array to group results by. 
               Alternatively, `index` can be used to use indexing arrays as the 
               basis for grouping.
    grouping_specifiers : dict, optional
                          dict of categories to be used as a basis for grouping 
                          by. Grouping_specifiers is only meaningful if 
                          group_by is provided as well. In case of grouping by
                          index, the grouping  specifiers should be in a 
                          dictionary where the key denotes the name of the 
                          group. 
    ylabels : dict, optional
              ylabels is a dictionary with the outcome names as keys, the 
              specified values will be used as labels for the y axis. 
    point_in_time : float, optional
                    the point in time at which the scatter is to be made. If 
                    None is provided (default), the end states are used. 
                    point_in_time should be a valid value on time
    log: bool, optional
        indicating whether density should be log scaled. Defaults to True.
    gridsize : int, optional
               controls the gridsize for the hexagonal bining. (default = 50)
    cmap : str
           color map that is to be used in generating the hexbin. For details 
           on the available maps, see `pylab <http://matplotlib.sourceforge.net/examples/pylab_examples/show_colormaps.html#pylab-examples-show-colormaps>`_.
           (Defaults = coolwarm)
    filter_scalar: bool, optional 
                   remove the non-time-series outcomes. Defaults to True.
    
    Returns
    -------
    fig
        the figure instance
    dict
        key is tuple of names of outcomes, value is associated axes
        instance
    
    '''
    debug("generating pairwise density plot")
    
    prepared_data = prepare_pairs_data(results, outcomes_to_show, group_by, 
                                       grouping_specifiers, point_in_time,
                                       filter_scalar)
    outcomes, outcomes_to_show, grouping_specifiers = prepared_data
   
    if group_by:
        #figure out the extents for each combination
        extents = determine_extents(outcomes, outcomes_to_show)
        
        axes_dicts = {}
        figures = []
        for key, value in outcomes.items():
            figure, axes_dict = simple_pairs_density(value, outcomes_to_show, 
                                       log, colormap, gridsize, ylabels,
                                       extents=extents, title=key)
            axes_dicts[key] = axes_dict
            figures.append(figure)
        
        # harmonize the color scaling across figures
        combis = [(field1, field2) for field1 in outcomes_to_show\
                           for field2 in outcomes_to_show]
        for combi in combis:
            vmax = -1
            for entry in axes_dicts.values():
                vmax =  max(entry[combi].collections[0].norm.vmax, vmax)
            for entry in axes_dicts.values():
                ax = entry[combi]
                ax.collections[0].set_clim(vmin=0, vmax=vmax)
            del vmax
            
        return figures, axes_dicts
    else:
        return simple_pairs_density(outcomes, outcomes_to_show, log, colormap, 
                                    gridsize, ylabels)


def determine_extents(outcomes, outcomes_to_show):
    '''
    Helper function used by pairs_density to make sure that multiple groups
    share the same axes extent.

    Parameters
    ----------
    outcomes : dict
    outcomes_to_show : list of str
    
    Returns
    -------
    dict
        tuple of str as key, and 4-tuple with extent
    
    '''
    
    limits = {}
    for pol_out in outcomes.values():
        for entry in outcomes_to_show:
            out = pol_out[entry]
            minimum = np.amin(out)
            maximum = np.amax(out)
            try:
                cur = limits[entry]
                new = (min(cur[0], minimum),
                     max(cur[1], maximum)) 
                limits[entry] = new
            except KeyError:
                limits[entry] = (minimum, maximum)
    extents = {}
    combis = [(field1, field2) for field1 in outcomes_to_show\
                               for field2 in outcomes_to_show]
    for field1, field2 in combis:
        limits_1 = limits[field1]
        limits_2 = limits[field2]  
        extents[(field1, field2)] = (limits_1[0], limits_1[1],
                                     limits_2[0], limits_2[1])
    return extents


def simple_pairs_density(outcomes,
                         outcomes_to_show,
                         log,
                         colormap,
                         gridsize,
                         ylabels,
                         extents=None,
                         title=None):
    '''
    
    Helper function for generating a simple pairs density plot

    Parameters
    ----------
    outcomes : dict
    outcomes_to_show : list of str
    log : bool
    colormap : str
    gridsize : int
    ylabels: dict
    extents : dict, optional
             used to control scaling of plots across figures, If provided, it 
             should be a dict with a tuple of outcomes as key and the extend to 
             be used as value.
    title : str, optional
                    
    
    '''
    
    grid = gridspec.GridSpec(len(outcomes_to_show), len(outcomes_to_show))
    grid.update(wspace = 0.1,
                hspace = 0.1)
     
    #the plotting
    figure = plt.figure()
     
    combis = [(field1, field2) for field1 in outcomes_to_show\
                               for field2 in outcomes_to_show]
    axes_dict = {}
    for field1, field2 in combis:
        i = list(outcomes_to_show).index(field1)
        j = list(outcomes_to_show).index(field2)
        
        ax = figure.add_subplot(grid[i,j])
        axes_dict[(field1, field2)] = ax

        y_data = outcomes[field1]
        x_data = outcomes[field2]

        bins=None
        if log:
            bins='log'
        
        extent = None
        if extents:
            extent = extents[(field2, field1)]
       
        #text and labels
        if i == j:
            #only plot the name in the middle
            ax.hexbin(x_data,y_data, bins=bins, gridsize=gridsize, 
                      cmap=cm.__dict__[colormap], alpha=0, edgecolor='white', 
                      linewidths=1, extent=extent)
        else:
            ax.hexbin(x_data,y_data, bins=bins, gridsize=gridsize, 
                      cmap=cm.__dict__[colormap], edgecolor='black', 
                      linewidths=1, extent=extent, mincnt=1)
        do_text_ticks_labels(ax, i, j, field1, field2, ylabels, 
                             outcomes_to_show)

    return figure, axes_dict

  
def pairs_scatter(results, 
                  outcomes_to_show = [],
                  group_by = None,
                  grouping_specifiers = None,
                  ylabels = {},
                  legend=True,
                  point_in_time=-1,
                  filter_scalar=False,
                  **kwargs):
    '''
    
    Generate a `R style pairs <http://www.stat.psu.edu/~dhunter/R/html/graphics/html/pairs.html>`_ 
    scatter multiplot. In case of time-series data, the end states are used.
    
    Parameters
    ----------
    results : tuple
              return from perform_experiments.
    outcomes_to_show : list of str, optional
                       list of outcome of interest you want to plot.
    group_by : str, optional
               name of the column in the cases array to group results by. 
               Alternatively, `index` can be used to use indexing arrays as the 
               basis for grouping.
    grouping_specifiers : dict, optional
                          dict of categories to be used as a basis for grouping 
                          by. Grouping_specifiers is only meaningful if 
                          group_by is provided as well. In case of grouping by
                          index, the grouping  specifiers should be in a 
                          dictionary where the key denotes the name of the 
                          group. 
    ylabels : dict, optional
              ylabels is a dictionary with the outcome names as keys, the 
              specified values will be used as labels for the y axis. 
    legend : bool, optional
             if true, and group_by is given, show a legend.
    point_in_time : float, optional
                    the point in time at which the scatter is to be made. If 
                    None is provided (default), the end states are used. 
                    point_in_time should be a valid value on time
    filter_scalar: bool, optional 
                   remove the non-time-series outcomes. Defaults to True.

    Returns
    -------
    fig : Figure instance
          the figure instance
    axes  : dict
            key is tuple of names of outcomes, value is associated axes
            instance

    .. note:: the current implementation is limited to seven different 
              categories in case of column, categories, and/or discretesize.
              This limit is due to the colors specified in COLOR_LIST. 
    
    '''
    
    debug("generating pairwise scatter plot")
   
    prepared_data = prepare_pairs_data(results, outcomes_to_show, group_by, 
                                       grouping_specifiers, point_in_time,
                                       filter_scalar)
    outcomes, outcomes_to_show, grouping_labels = prepared_data
   
    grid = gridspec.GridSpec(len(outcomes_to_show), len(outcomes_to_show))                             
    grid.update(wspace = 0.1,
                hspace = 0.1)    
    
    
    #the plotting
    figure = plt.figure()
    axes_dict = {}
    
    combis = [(field1, field2) for field1 in outcomes_to_show\
                               for field2 in outcomes_to_show]
    
    for field1, field2 in combis:
        i = list(outcomes_to_show).index(field1)
        j = list(outcomes_to_show).index(field2)
        ax = figure.add_subplot(grid[i,j])
        axes_dict[(field1, field2)] = ax

        if group_by:
            for x, group in enumerate(grouping_labels):
                y_data = outcomes[group][field1]
                x_data = outcomes[group][field2]
                
                facecolor = plotting_util.COLOR_LIST[x]
                edgecolor = 'k'
                if i==j: 
                    facecolor = 'white'
                    edgecolor = 'white'
                ax.scatter(x_data, y_data, 
                           facecolor=facecolor, edgecolor=edgecolor)
        else:
            y_data = outcomes[field1]
            x_data = outcomes[field2]

            facecolor = 'b'
            edgecolor = 'k'
            if i==j: 
                facecolor = 'white'
                edgecolor = 'white'
            ax.scatter(x_data, y_data, 
                       facecolor=facecolor, edgecolor=edgecolor)
        do_text_ticks_labels(ax, i, j, field1, field2, ylabels, 
                             outcomes_to_show)

    if group_by and legend:
        gs1 = grid[0,0]
        
        for ax in figure.axes:
            gs2 = ax._subplotspec
            if all((gs1._gridspec == gs2._gridspec,
                    gs1.num1 == gs2.num1,
                    gs1.num2 == gs2.num2)):
                break  
        
        make_legend(grouping_labels, ax, legend_type=SCATTER)

    return figure, axes_dict


def do_text_ticks_labels(ax, i, j, field1, field2, ylabels, outcomes_to_show):
    '''
    
    Helper function for setting the tick labels on the axes correctly on and 
    off

    Parameters
    ----------
    ax : axes
    i : int
    j : int
    field1 : str
    field2 : str
    ylabels : dict, optional
    outcomes_to_show : str
    
    
    '''
    
    #text and labels
    if i == j:
        #only plot the name in the middle
        if ylabels:
            text = ylabels[field1]
        else:
            text = field1
        ax.text(0.5, 0.5, text,
                horizontalalignment='center',
                verticalalignment='center',
                transform = ax.transAxes)  
    
    # are we at the end of the row?
    if i != len(outcomes_to_show)-1:
        #xaxis off
        ax.set_xticklabels([])
    else:
        if ylabels:
            try:
                ax.set_xlabel(ylabels.get(field2))
            except KeyError:
                info("no label specified for "+field2)
        else:
            ax.set_xlabel(field2) 
    
    # are we at the end of the column?
    if j != 0:
        #yaxis off
        ax.set_yticklabels([])
    else:
        if ylabels:
            try:
                ax.set_ylabel(ylabels.get(field1))
            except KeyError:
                info("no label specified for "+field1) 
        else:
            ax.set_ylabel(field1)   



'''
Created on 20 dec. 2010

This file illustrated the use the EMA classes for a contrived example
It's main purpose has been to test the parallel processing functionality

.. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>
'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)

from ema_workbench import (Model, RealParameter, ScalarOutcome, ema_logging,
                           perform_experiments)

def some_model(x1=None, x2=None, x3=None):
    return {'y':x1*x2+x3}

if __name__ == '__main__':
    ema_logging.LOG_FORMAT = '[%(name)s/%(levelname)s/%(processName)s] %(message)s'
    ema_logging.log_to_stderr(ema_logging.INFO)
    
    model = Model('simpleModel', function=some_model) #instantiate the model

    #specify uncertainties
    model.uncertainties = [RealParameter("x1", 0.1, 10),
                           RealParameter("x2", -0.01,0.01),
                           RealParameter("x3", -0.01,0.01)]
    #specify outcomes 
    model.outcomes = [ScalarOutcome('y')]

    results = perform_experiments(model, 100)



from __future__ import (absolute_import, unicode_literals, division, 
                        print_function)


__all__ = ["ema_parallel", "parameters"
           "model", "outcomes", "samplers", 
           "Model", 'FileModel', "ModelEnsemble",
           "Outcome", "ScalarOutcome", "TimeSeriesOutcome",
           "RealParameter", "IntegerParameter", "CategoricalParameter",
           "Scenario", "Policy", "Experiment", "Constant", "create_parameters",
           "parameters_to_csv", "Category", "SobolSampler", "MorrisSampler",
           "get_SALib_problem", "FASTSampler"
           "peform_experiments", 'optimize', "IpyparallelEvaluator", 
           "MultiprocessingEvaluator", "SequentialEvaluator"
           ]

from .outcomes import ScalarOutcome, TimeSeriesOutcome, Outcome
from .model import Model, FileModel
from .parameters import (RealParameter, IntegerParameter, CategoricalParameter,
                     Scenario, Policy, Constant, Experiment, create_parameters,
                     parameters_to_csv, Category, experiment_generator)
from .samplers import (MonteCarloSampler, FullFactorialSampler, LHSSampler, 
                       PartialFactorialSampler, sample_levers, 
                       sample_uncertainties)
from .salib_samplers import (SobolSampler, MorrisSampler, FASTSampler, 
                             get_SALib_problem)
from .evaluators import (perform_experiments, optimize, IpyparallelEvaluator, 
                         MultiprocessingEvaluator, SequentialEvaluator)



'''


'''
from __future__ import (unicode_literals, print_function, absolute_import,
                                        division)

import pandas as pd
import warnings

from .outcomes import AbstractOutcome
from .parameters import (IntegerParameter, RealParameter, CategoricalParameter,
                         Scenario, Policy)
from .samplers import determine_parameters
from .util import determine_objects

try:
    from platypus import EpsNSGAII  # @UnresolvedImport
    from platypus import Problem as PlatypusProblem
    import platypus
except ImportError:
    warnings.warn("platypus based optimization not available", ImportWarning)
    class PlatypusProblem(object):
        def __init__(self, *args, **kwargs):
            pass 
    EpsNSGAII = None
    platypus = None

# Created on 5 Jun 2017
#
# .. codeauthor::jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

__all__ = []


class Problem(PlatypusProblem):
    '''small extension to Platypus problem object, includes information on
    the names of the decision variables, the names of the outcomes,
    and the type of search'''
    
    def __init__(self, searchover, parameters, parameter_names,
                 outcome_names):
        super(Problem, self).__init__(len(parameters), len(outcome_names) , 
                                      nconstrs=0)
        assert len(parameters) == len(parameter_names)
        assert searchover in ('levers', 'uncertainties', 'robust')
        
        self.searchover = searchover
        self.parameters = parameters
        self.parameter_names = parameter_names
        self.outcome_names = outcome_names


class RobustProblem(Problem):
    '''small extension to Problem object for robust optimization, adds the 
    scenarios and the robustness functions'''
    
    
    def __init__(self, parameters, parameter_names,
                 outcome_names, scenarios, robustness_functions):
        super(RobustProblem, self).__init__('robust', parameters, 
                                            parameter_names, outcome_names)
        assert len(robustness_functions) == len(outcome_names)        
        self.scenarios = scenarios
        self.robustness_functions = robustness_functions


def to_problem(model, searchover):
    '''helper function to create Problem object
    
    Parameters
    ----------
    model : AbstractModel instance
    searchover : str
    
    Returns
    -------
    Problem instance
    
    '''
    _type_mapping = {RealParameter: platypus.Real,
                     IntegerParameter: platypus.Integer,
                     CategoricalParameter: platypus.Permutation}
    
    # extract the levers and the outcomes
    decision_variables = determine_parameters(model, searchover, union=True)
    dvnames = [dv.name for dv in decision_variables]

    outcomes = determine_objects(model,'outcomes')
    outcomes = [outcome for outcome in outcomes if 
                outcome.kind != AbstractOutcome.INFO]
    outcome_names = [outcome.name for outcome in outcomes]
    
    
    problem = Problem(searchover, decision_variables, dvnames,
                      outcome_names)
    problem.types = to_platypus_types(decision_variables)
    problem.directions = [outcome.kind for outcome in outcomes]

    return problem



def to_robust_problem(model, scenarios, robustness_functions):
    '''helper function to create RobustProblem object
    
    Parameters
    ----------
    model : AbstractModel instance
    robustness_functions : iterable of ScalarOutcomes
    
    Returns
    -------
    RobustProblem instance
    
    '''

    
    # extract the levers and the outcomes
    decision_variables = determine_parameters(model, 'levers', union=True)
    dvnames = [dv.name for dv in decision_variables]

    outcomes = robustness_functions
    outcome_names = [outcome.name for outcome in outcomes]
    
    problem = RobustProblem(decision_variables, dvnames, outcome_names, 
                            scenarios, robustness_functions)
    
    problem.types = to_platypus_types(decision_variables)
    problem.directions = [outcome.kind for outcome in outcomes]

    return problem


def to_platypus_types(decision_variables):
    
    _type_mapping = {RealParameter: platypus.Real,
                     IntegerParameter: platypus.Integer,
                     CategoricalParameter: platypus.Permutation}
    types = []
    for dv in decision_variables:
        klass = _type_mapping[type(dv)]
        
        if not isinstance(dv, CategoricalParameter):
            decision_variable = klass(dv.lower_bound, dv.upper_bound)
        else:
            decision_variable = klass(dv.categories)
        
        types.append(decision_variable)
    return types


def to_dataframe(optimizer, dvnames, outcome_names):
    '''helper function to turn results of optimization into a pandas DataFrame
    
    Parameters
    ----------
    optimizer : platypus algorithm instance
    dvnames : list of str
    outcome_names : list of str
    
    Returns
    -------
    pandas DataFrame
    
    
    '''
    
    solutions = []
    for solution in platypus.unique(platypus.nondominated(optimizer.result)):
        decision_vars = dict(zip(dvnames, solution.variables))
        decision_out = dict(zip(outcome_names, solution.objectives))
        
        result = decision_vars.copy()
        result.update(decision_out)
        
        solutions.append(result)

    results = pd.DataFrame(solutions, columns=dvnames+outcome_names)
    return results


def process_uncertainties(jobs):
    scenarios = []
    policies = 0
    
    for i, platypus_job in enumerate(jobs):
        variables = dict(zip(platypus_job.solution.problem.parameter_names, 
                             platypus_job.solution.variables))
        name = str(i)
        job = Scenario(name=name, **variables)
        scenarios.append(job)
        
    return scenarios, policies

def process_levers(jobs):
    policies = []
    scenarios = 0
    
    for i, platypus_job in enumerate(jobs):
        variables = dict(zip(platypus_job.solution.problem.parameter_names, 
                             platypus_job.solution.variables))
        name = str(i)
        job = Policy(name=name, **variables)
        policies.append(job)
    
    return scenarios, policies


def process_robust(jobs):
    _, policies = process_levers(jobs)
    scenarios = jobs[0].solution.problem.scenarios
    
    return scenarios, policies


def evaluate(jobs_collection, experiments, outcomes, problem):
    
    searchover = problem.searchover
    outcome_names = problem.outcome_names
    
    if searchover=='levers':
        column = 'policy'
    else:
        column = 'scenario_id'
    
    for entry, job in jobs_collection:
        logical = experiments[column] == entry.name
        job_outcomes = [outcomes[key][logical][0] for key in outcome_names]
        
        job.solution.problem.function = lambda x: job_outcomes
        job.solution.evaluate()


def evaluate_robust(jobs_collection, experiments, outcomes,
                     problem):
    robustness_functions = problem.robustness_functions
    
    for entry, job in jobs_collection:
        logical = experiments['policy'] == entry.name

        job_outcomes = []
        for rf in robustness_functions:
            data = [outcomes[var_name][logical] for var_name in 
                    rf.variable_name]
            job_outcomes.append(rf.function(*data))
        
        job.solution.problem.function = lambda x: job_outcomes
        job.solution.evaluate()


'''


'''
from __future__ import (unicode_literals, print_function, absolute_import,
                                        division)

import math
import numpy as np
    
from scipy.optimize import brentq

from ema_workbench import (Model, RealParameter, ScalarOutcome, Constant, 
                           ema_logging, MultiprocessingEvaluator)
from ema_workbench.em_framework.samplers import sample_uncertainties


# Created on 1 Jun 2017
#
# .. codeauthor::jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

__all__ = []

def get_antropogenic_release(xt, c1, c2, r1, r2, w1):
    '''
    
    Parameters
    ----------
    xt : float
         polution in lake at time t
    c1 : float
         center rbf 1
    c2 : float
         center rbf 2
    r1 : float
         ratius rbf 1
    r2 : float
         ratius rbf 2
    w1 : float
         weight of rbf 1
         
    note:: w2 = 1 - w1
    
    '''
    
    rule = w1*(abs(xt-c1/r1))**3+(1-w1)*(abs(xt-c2/r2)**3)
    at1 = max(rule, 0.01)
    at = min(at1, 0.1)
    
    return at


def lake_problem(
         b=0.42,          # decay rate for P in lake (0.42 = irreversible)
         q=2.0,           # recycling exponent
         mean=0.02,       # mean of natural inflows
         stdev=0.001,     # future utility discount rate
         delta=0.98,      # standard deviation of natural inflows
         alpha=0.4,       # utility from pollution
         nsamples=100,    # Monte Carlo sampling of natural inflows
         myears=1,         # the runtime of the simulation model
         c1=0.25,
         c2=0.25,
         r1=0.5,
         r2=0.5,
         w1=0.5):   
    Pcrit = brentq(lambda x: x**q/(1+x**q) - b*x, 0.01, 1.5)

    X = np.zeros((myears,))
    average_daily_P = np.zeros((myears,))
    reliability = 0.0
    inertia = 0
    utility = 0

    for _ in range(nsamples):
        X[0] = 0.0
        decision = 0.1

        decisions = np.zeros(myears,)
        decisions[0] = decision

        natural_inflows = np.random.lognormal(
                math.log(mean**2 / math.sqrt(stdev**2 + mean**2)),
                math.sqrt(math.log(1.0 + stdev**2 / mean**2)),
                size=myears)
        
        for t in range(1,myears):
        
            # here we use the decision rule
            decision = get_antropogenic_release(X[t-1], c1, c2, r1, r2, w1)
            decisions[t] = decision
               
            X[t] = (1-b)*X[t-1] + X[t-1]**q/(1+X[t-1]**q) + decision +\
                    natural_inflows[t-1]
            average_daily_P[t] += X[t]/nsamples
    
        reliability += np.sum(X < Pcrit)/(nsamples*myears)
        inertia += np.sum(np.absolute(np.diff(decisions) > 0.02)) / (nsamples*myears)
        utility += np.sum(alpha*decisions*np.power(delta, np.arange(myears))) / nsamples
    max_P = np.max(average_daily_P)

    return max_P, utility, inertia, reliability

if __name__ == '__main__':
    
    ema_logging.log_to_stderr(ema_logging.INFO)

    #instantiate the model
    lake_model = Model('lakeproblem', function=lake_problem)
    #specify uncertainties
    lake_model.uncertainties = [RealParameter('b', 0.1, 0.45),
                                RealParameter('q', 2.0, 4.5),
                                RealParameter('mean', 0.01, 0.05),
                                RealParameter('stdev', 0.001, 0.005),
                                RealParameter('delta', 0.93, 0.99)]
    
    # set levers
    lake_model.levers = [RealParameter("c1", -2, 2),
                         RealParameter("c2", -2, 2),
                         RealParameter("r1", 0, 2), 
                         RealParameter("r2", 0, 2), 
                         RealParameter("w1", 0, 1)
                         ]
    
    #specify outcomes 
    lake_model.outcomes = [ScalarOutcome('max_P'),
                           ScalarOutcome('utility'), 
                           ScalarOutcome('inertia'), 
                           ScalarOutcome('reliability')]  
    
    # override some of the defaults of the model
    lake_model.constants = [Constant('alpha', 0.41),
                            Constant('nsamples', 100),
                            Constant('myears', 100)]
    
    def signal_to_noise(data):
        mean = np.mean(data)
        std = np.std(data)
        sn = mean/std
        return sn
    
    
    MAXIMIZE = ScalarOutcome.MAXIMIZE  # @UndefinedVariable
    MINIMIZE = ScalarOutcome.MINIMIZE  # @UndefinedVariable
    robustnes_functions = [ScalarOutcome('mean p', kind=MINIMIZE, 
                                 variable_name='max_P', function=signal_to_noise),
                           ScalarOutcome('std p', kind=MINIMIZE, 
                                         variable_name='max_P', function=np.std)]
    n_scenarios = 10
    scenarios = sample_uncertainties(lake_model, n_scenarios)
    nfe = 1000
    
    with MultiprocessingEvaluator(lake_model) as evaluator:
        evaluator.robust_optimize(robustnes_functions, scenarios, nfe=nfe, 
                                epsilons=[0.1,]*len(robustnes_functions),
                                population_size=25)



'''


'''
from __future__ import (unicode_literals, print_function, absolute_import,
                                        division)



# Created on Jul 23, 2016
#
# .. codeauthor::jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

from ema_workbench import (RealParameter, TimeSeriesOutcome,ema_logging,
                           perform_experiments)

from ema_workbench.connectors.pysd_connector import PysdModel

if __name__ == '__main__':
    ema_logging.log_to_stderr(ema_logging.DEBUG)    
    
    mdl_file = './models/pysd/Teacup.mdl'
    
    model = PysdModel(mdl_file=mdl_file)
    
    model.uncertainties = [RealParameter('Room Temperature', 33, 120)]
    model.outcomes = [TimeSeriesOutcome('Teacup Temperature')]
    
    perform_experiments(model, 100)


from __future__ import absolute_import

# from . import prim
# from . import cart
# from . import b_and_w_plotting
# from . import plotting
# from . import plotting_util
# from . import pairs_plotting
# from . import feature_scoring 


"""


this module provides functions for generating some basic figures. The code can
be used as is, or serve as an example for writing your own code. 

"""
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)

import copy

import matplotlib.pyplot as plt
import numpy as np
import six
from matplotlib.patches import ConnectionPatch

from . import plotting_util
from .plotting_util import (prepare_data, simple_kde, group_density, make_grid,
                           make_legend, plot_envelope, simple_density,
                           do_titles, do_ylabels, TIME, ENV_LIN, ENVELOPE,
                           LINES, PATCH, LINE, TIGHT)
from ..util import debug, warning, EMAError

# .. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

__all__ = ['lines', 
           'envelopes', 
           'kde_over_time', 
           'multiple_densities']

TIME_LABEL = 'Time'


def envelopes(results, 
              outcomes_to_show = [],
              group_by = None,
              grouping_specifiers = None,
              density=None,
              fill=False,
              legend=True,
              titles={},
              ylabels={},
              log=False):
    ''' Make envelop plots. An envelope shows over time the minimum and maximum 
    value for a set of runs over time. It is thus to be used in case of time 
    series data. The function will try to find a result labeled "TIME". If this
    is present, these values will be used on the X-axis. In case of Vensim 
    models, TIME is present by default.

    Parameters
    ----------
    results : tupule
              return from :meth:`perform_experiments`.
    outcomes_to_show : list of str, optional
                       list of outcome of interest you want to plot. If empty, 
                       all outcomes are plotted. **Note**:  just names.
    group_by : str, optional
               name of the column in the cases array to group results by. 
               Alternatively, `index` can be used to use indexing arrays as the
               basis for grouping.
    grouping_specifiers : iterable or dict, optional
                          set of categories to be used as a basis for grouping 
                          by. Grouping_specifiers is only meaningful if 
                          group_by is provided as well. In case of grouping by 
                          index, the grouping specifiers should be in a 
                          dictionary where the key denotes the name of the 
                          group. 
    density : {None, HIST, KDE, VIOLIN, BOXPLOT}, optional
    fill : bool, optional
    legend : bool, optional
    titles : dict, optional
             a way for controlling whether each of the axes should have a 
             title. There are three possibilities. If set to None, no title 
             will be shown for any of the axes. If set to an empty dict, 
             the default, the title is identical to the name of the outcome of 
             interest. If you want to override these default names, provide a 
             dict with the outcome of interest as key and the desired title as 
             value. This dict need only contain the outcomes for which you 
             want to use a different title. 
    ylabels : dict, optional
              way for controlling the ylabels. Works identical to titles.
    log : bool, optional
          log scale density plot

    Returns
    -------
    Figure : Figure instance
    axes : dict
           dict with outcome as key, and axes as value. Density axes' are
           indexed by the outcome followed by _density.
        
    Note
    ----
    the current implementation is limited to seven different categories in case 
    of group_by, categories, and/or discretesize. This limit is due to the colors 
    specified in COLOR_LIST.
    
    Examples
    --------

    >>> import util as util
    >>> data = util.load_results(r'1000 flu cases.cPickle')
    >>> envelopes(data, group_by='policy')

    will show an envelope for three three different policies, for all the 
    outcomes of interest. while

    >>> envelopes(data, group_by='policy', categories=['static policy', 
                  'adaptive policy'])

    will only show results for the two specified policies, ignoring any results 
    associated with \'no policy\'.

    '''
    debug("generating envelopes")
    results = copy.deepcopy(results)
   
    prepared_data = prepare_data(results, outcomes_to_show, group_by,
                                 grouping_specifiers)
    outcomes, outcomes_to_show, time, grouping_labels = prepared_data
    
    figure, grid = make_grid(outcomes_to_show, density)
    
    # do the plotting
    axes_dict = {}
    for i, outcome_to_plot in enumerate(outcomes_to_show):
        ax = figure.add_subplot(grid[i,0])
        axes_dict[outcome_to_plot] = ax
        
        ax_d= None
        if density:
            ax_d = figure.add_subplot(grid[i,1], sharey=ax)
            axes_dict[outcome_to_plot+"_density"] = ax_d
    
        if group_by:
            group_by_envelopes(outcomes,outcome_to_plot, time, density,
                               ax, ax_d, fill, grouping_labels, log)
        else:
            single_envelope(outcomes, outcome_to_plot, time, density,
                            ax, ax_d, fill, log)
            
        if ax_d:
            for tl in ax_d.get_yticklabels():
                tl.set_visible(False)
        
        ax.set_xlabel(TIME_LABEL)
        do_ylabels(ax, ylabels, outcome_to_plot)
        do_titles(ax, titles, outcome_to_plot)

    if legend and group_by:
        gs1 = grid[0,0]
        
        for ax in figure.axes:
            gs2 = ax._subplotspec
            if all((gs1._gridspec == gs2._gridspec,
                    gs1.num1 == gs2.num1,
                    gs1.num2 == gs2.num2)):
                break
        if fill:
            make_legend(grouping_labels, ax, alpha=0.3, 
                        legend_type=PATCH)
        else:
            make_legend(grouping_labels, ax, legend_type=LINE)
    
    if TIGHT:
        grid.tight_layout(figure)
    
    return figure, axes_dict


def group_by_envelopes(outcomes,
                       outcome_to_plot,
                       time,
                       density,
                       ax,
                       ax_d,
                       fill,
                       group_labels, 
                       log):
    ''' Helper function, responsible for generating an envelope plot based on
    a grouping. 

    Parameters
    ----------
    outcomes : dict
               a dictonary containing the various outcomes to plot
    outcome_to_plot : str
                      the specific outcome to plot
    time : str
           the name of the time dimension
    density :  {None, HIST, KDE, VIOLIN, BOXPLOT} 
    ax : Axes instance
         the ax on which to plot
    ax_d : Axes instance
           the ax on which to plot the density
    fill : bool
    group_by_labels : list of str
                      order in which groups should be plotted
    log : bool

    '''
    
    for j, key in enumerate(group_labels):
        value = outcomes[key]
        value = value[outcome_to_plot]
        try:
            plot_envelope(ax, j, time, value,fill)
        except ValueError:
            warning("value error when plotting for %s" % (key))
            raise
    
    if density:
        group_density(ax_d, density, outcomes, outcome_to_plot, group_labels, 
                      log)
    
        ax_d.get_yaxis().set_view_interval(
                     ax.get_yaxis().get_view_interval()[0],
                     ax.get_yaxis().get_view_interval()[1])


def single_envelope(outcomes,
                    outcome_to_plot,
                    time,
                    density,
                    ax,
                    ax_d,
                    fill, 
                    log):
    '''

    Helper function for generating a single envelope plot.

    Parameters
    ----------
    outcomes : dict
               a dictonary containing the various outcomes to plot
    outcome_to_plot : str
                      the specific outcome to plot
    time : str
           the name of the time dimension
    density :  {None, HIST, KDE, VIOLIN, BOXPLOT} 
    ax : Axes instance
         the ax on which to plot
    ax_d : Axes instance
           the ax on which to plot the density
    fill : bool
    group_by_labels : list of str
                      order in which groups should be plotted
    log : bool

    '''
    value = outcomes[outcome_to_plot]
    
    plot_envelope(ax, 0, time, value,fill)
    if density:
        simple_density(density, value, ax_d, ax, log)

          
def lines(results, 
          outcomes_to_show = [],
          group_by = None,
          grouping_specifiers = None,
          density='',
          legend=True,
          titles={},
          ylabels={},
          experiments_to_show=None,
          show_envelope=False,
          log=False):
    '''This function takes the results from :meth:`perform_experiments` and 
    visualizes these as line plots. It is thus to be used in case of time 
    series data. The function will try to find a result labeled "TIME". If this
    is present, these values will be used on the X-axis. In case of Vensim 
    models, TIME is present by default.  

    Parameters
    ----------
    results : tuple
              return from :meth:`perform_experiments`.
    outcomes_to_show : list of str, optional
                       list of outcome of interest you want to plot. If empty, 
                       all outcomes are plotted. **Note**:  just names.
    group_by : str, optional
               name of the column in the cases array to group results by. 
               Alternatively, `index` can be used to use indexing arrays as the
               basis for grouping.
    grouping_specifiers : iterable or dict, optional
                          set of categories to be used as a basis for grouping 
                          by. Grouping_specifiers is only meaningful if 
                          group_by is provided as well. In case of grouping by 
                          index, the grouping specifiers should be in a 
                          dictionary where the key denotes the name of the 
                          group. 
    density : {None, HIST, KDE, VIOLIN, BOXPLOT}, optional
    legend : bool, optional
    titles : dict, optional
             a way for controlling whether each of the axes should have a 
             title. There are three possibilities. If set to None, no title 
             will be shown for any of the axes. If set to an empty dict, 
             the default, the title is identical to the name of the outcome of 
             interest. If you want to override these default names, provide a 
             dict with the outcome of interest as key and the desired title as 
             value. This dict need only contain the outcomes for which you 
             want to use a different title. 
    ylabels : dict, optional
              way for controlling the ylabels. Works identical to titles.
    experiments_to_show : ndarray, optional
                          indices of experiments to show lines for,
                          defaults to None.
    show_envelope : bool, optional
                    show envelope op outcomes. This envelope is the based on
                    the minimum at each column and the maximum at each column.
    log : bool, optional
          log scale density plot

    Returns
    -------
    fig : Figure instance
    axes : dict
           dict with outcome as key, and axes as value. Density axes' are
           indexed by the outcome followed by _density.
    
    Note
    ----
    the current implementation is limited to seven different categories in case 
    of group_by, categories, and/or discretesize. This limit is due to the colors 
    specified in COLOR_LIST.

    '''
    
    debug("generating line graph")

    # make sure we have the data
    results = copy.deepcopy(results)

    if show_envelope:
        return plot_lines_with_envelopes(results, 
                                outcomes_to_show=outcomes_to_show, 
                                group_by=group_by, legend=legend, 
                                density=density, 
                                grouping_specifiers=grouping_specifiers, 
                                experiments_to_show=experiments_to_show, 
                                titles=titles, ylabels=ylabels, log=log)
    
    if experiments_to_show is not None:
        experiments, outcomes = results
        experiments = experiments[experiments_to_show]
        new_outcomes = {}
        for key, value in outcomes.items():
            new_outcomes[key] = value[experiments_to_show]
        results = experiments, new_outcomes

    data = prepare_data(results, outcomes_to_show, group_by, 
                        grouping_specifiers)
    outcomes, outcomes_to_show, time, grouping_labels = data

    figure, grid = make_grid(outcomes_to_show, density)
    axes_dict = {}

    # do the plotting
    for i, outcome_to_plot in enumerate(outcomes_to_show):
        ax = figure.add_subplot(grid[i,0])
        axes_dict[outcome_to_plot] = ax
        
        ax_d= None
        if density:
            ax_d = figure.add_subplot(grid[i,1], sharey=ax)
            axes_dict[outcome_to_plot+"_density"] = ax_d
            
            for tl in ax_d.get_yticklabels():
                tl.set_visible(False)
    
        if group_by:
            group_by_lines(outcomes,outcome_to_plot, time, density,
                           ax, ax_d, grouping_labels, log)
        else:
            simple_lines(outcomes, outcome_to_plot, time, density,
                         ax, ax_d, log)
        ax.set_xlabel(TIME_LABEL)
        do_ylabels(ax, ylabels, outcome_to_plot)
        do_titles(ax, titles, outcome_to_plot)
            
    if legend and group_by:
        gs1 = grid[0,0]
        
        for ax in figure.axes:
            gs2 = ax._subplotspec
            if all((gs1._gridspec == gs2._gridspec,
                    gs1.num1 == gs2.num1,
                    gs1.num2 == gs2.num2)):
                break        
        
        make_legend(grouping_labels, ax)
    
    if TIGHT:
        grid.tight_layout(figure)
    
    return figure, axes_dict


def plot_lines_with_envelopes(results, 
                              outcomes_to_show = [],
                              group_by = None,
                              grouping_specifiers = None,
                              density='',
                              legend=True,
                              titles={},
                              ylabels={},
                              experiments_to_show=None,
                              log=False):
    '''

    Helper function for generating a plot which contains both an envelope and
    lines.  

    Parameters
    ----------
    results : tupule
              return from :meth:`perform_experiments`.
    outcomes_to_show : list of str, optional
                       list of outcome of interest you want to plot. If empty, 
                       all outcomes are plotted. **Note**:  just names.
    group_by : str, optional
               name of the column in the cases array to group results by. 
               Alternatively, `index` can be used to use indexing arrays as the
               basis for grouping.
    grouping_specifiers : iterable or dict, optional
                          set of categories to be used as a basis for grouping 
                          by. Grouping_specifiers is only meaningful if 
                          group_by is provided as well. In case of grouping by 
                          index, the grouping specifiers should be in a 
                          dictionary where the key denotes the name of the 
                          group. 
    density : {None, HIST, KDE, VIOLIN, BOXPLOT}, optional
    legend : bool, optional
    titles : dict, optional
             a way for controlling whether each of the axes should have a 
             title. There are three possibilities. If set to None, no title 
             will be shown for any of the axes. If set to an empty dict, 
             the default, the title is identical to the name of the outcome of 
             interest. If you want to override these default names, provide a 
             dict with the outcome of interest as key and the desired title as 
             value. This dict need only contain the outcomes for which you 
             want to use a different title. 
    ylabels : dict, optional
              way for controlling the ylabels. Works identical to titles.
    experiments_to_show : ndarray, optional
                          indices of experiments to show lines for,
                          defaults to None.
    log : bool, optional

    Returns
    -------
    Figure
        a figure instance
    dict
        dict with outcome as key, and axes as value. Density axes' are
        indexed by the outcome followed by _density
    '''
   
    full_outcomes = prepare_data(results, outcomes_to_show, group_by,
                             grouping_specifiers)[0]
   
    experiments, outcomes = results
    experiments = experiments[experiments_to_show]
    new_outcomes={}
    for key, value in outcomes.items():
        new_outcomes[key] = value[experiments_to_show]
    results = experiments, new_outcomes

    data = prepare_data(results, outcomes_to_show, group_by, 
                        grouping_specifiers)
    outcomes, outcomes_to_show, time, grouping_labels = data
    


    figure, grid = make_grid(outcomes_to_show, density)
    axes_dict = {}

    # do the plotting
    for i, outcome_to_plot in enumerate(outcomes_to_show):
        ax = figure.add_subplot(grid[i,0])
        axes_dict[outcome_to_plot] = ax
        
        ax_d= None
        if density:
            ax_d = figure.add_subplot(grid[i,1], sharey=ax)
            axes_dict[outcome_to_plot+"_density"] = ax_d
            
            for tl in ax_d.get_yticklabels():
                tl.set_visible(False)
    
        if group_by:
            for j, key in enumerate(grouping_labels):
                full_value = full_outcomes[key][outcome_to_plot]
                plot_envelope(ax, j, time, full_value, fill=True)
            for j, key in enumerate(grouping_labels):
                value = outcomes[key][outcome_to_plot]
                full_value = full_outcomes[key][outcome_to_plot]
                ax.plot(time.T[:, np.newaxis], value.T, 
                        c=plotting_util.COLOR_LIST[j])
            
            if density:
                group_density(ax_d, density, full_outcomes, outcome_to_plot, 
                              grouping_labels, log)
               
                ax_d.get_yaxis().set_view_interval(
                             ax.get_yaxis().get_view_interval()[0],
                             ax.get_yaxis().get_view_interval()[1])
            
        else:
            value = full_outcomes[outcome_to_plot]
            plot_envelope(ax, 0, time, value, fill=True)
            if density:
                simple_density(density, value, ax_d, ax, log)
            
            value = outcomes[outcome_to_plot]
            ax.plot(time.T, value.T)
        
        ax.set_xlim(xmin=time[0] , xmax=time[-1])
        ax.set_xlabel(TIME_LABEL)
        do_ylabels(ax, ylabels, outcome_to_plot)
        do_titles(ax, titles, outcome_to_plot)

    if legend and group_by:
        gs1 = grid[0,0]
        
        for ax in figure.axes:
            gs2 = ax._subplotspec
            if all((gs1._gridspec == gs2._gridspec,
                    gs1.num1 == gs2.num1,
                    gs1.num2 == gs2.num2)):
                break
        make_legend(grouping_labels, ax)
    
    if TIGHT:
        grid.tight_layout(figure)
    
    return figure, axes_dict


def group_by_lines(outcomes, outcome_to_plot, time, density,
                   ax, ax_d, group_by_labels, log):
    '''

    Helper function responsible for generating a grouped lines plot. 
 
    Parameters
    ----------
    results : tupule
              return from :meth:`perform_experiments`.
    outcome_to_plot : str
    time : str
    density : {None, HIST, KDE, VIOLIN, BOXPLOT}
    ax : Axes instance
    ax_d : Axes instance
    group_by_labels : list of str
    log : bool

    '''
    
    for j, key in enumerate(group_by_labels):
        value = outcomes[key]
        value = value[outcome_to_plot]

        color = plotting_util.COLOR_LIST[j]
        ax.plot(time.T[:, np.newaxis], value.T, c=color, ms=1, markevery=5)
    
    if density:
        group_density(ax_d, density, outcomes, outcome_to_plot, 
                      group_by_labels, log)
        
        ax_d.get_yaxis().set_view_interval(
                     ax.get_yaxis().get_view_interval()[0],
                     ax.get_yaxis().get_view_interval()[1])


def simple_lines(outcomes, outcome_to_plot, time, density,
                 ax, ax_d, log):
    '''

    Helper function responsible for generating a simple lines plot. 

    Parameters
    ----------
    results : tupule
              return from :meth:`perform_experiments`.
    outcomes_to_plot : str
    time : str
    density : {None, HIST, KDE, VIOLIN, BOXPLOT}
    ax : Axes instance
    ax_d : Axes instance
    log : bool

    '''    
    value = outcomes[outcome_to_plot]
    ax.plot(time.T, value.T)
    if density:
        simple_density(density, value, ax_d, ax, log)


def kde_over_time(results, 
                  outcomes_to_show = [],
                  group_by = None,
                  grouping_specifiers = None,
                  results_to_show=None,
                  colormap='coolwarm',
#                  color_bar=False,
                  log=True):
    '''

    Plot a KDE over time. The KDE is is visualized through a heatmap

    Parameters
    ----------
    results : tupule
              return from :meth:`perform_experiments`.
    outcomes_to_show : list of str, optional
                       list of outcome of interest you want to plot. If empty, 
                       all outcomes are plotted. **Note**:  just names.
    group_by : str, optional
               name of the column in the cases array to group results by. 
               Alternatively, `index` can be used to use indexing arrays as the
               basis for grouping.
    grouping_specifiers : iterable or dict, optional
                          set of categories to be used as a basis for grouping 
                          by. Grouping_specifiers is only meaningful if 
                          group_by is provided as well. In case of grouping by 
                          index, the grouping specifiers should be in a 
                          dictionary where the key denotes the name of the 
                          group. 
    colormap : str, optional
               valid matplotlib color map name
    log : bool, optional

    Returns
    -------
    list of Figure instances
        a figure instance for each group for each outcome
    dict
        dict with outcome as key, and axes as value. Density axes' are
        indexed by the outcome followed by _density
       
    '''
    # TODO:: a colorbar boolean should be added. This controls whether a
    #        colorbar is shown for each axes.
    results = copy.deepcopy(results)

    #determine the minima and maxima over all runs
    minima = {}
    maxima = {}
    for key, value in results[1].items():
        minima[key] = np.min(value)
        maxima[key] = np.max(value)

    prepared_data = prepare_data(results, outcomes_to_show, group_by, 
                                 grouping_specifiers)
    outcomes, outcomes_to_show, time, grouping_specifiers = prepared_data
    del time

    if group_by:
        figures = []
        axes_dicts = {}
        for key, value in outcomes.items():
            fig, axes_dict = simple_kde(value, outcomes_to_show, colormap, log, 
                                        minima, maxima)
            fig.suptitle(key)
            figures.append(fig)
            axes_dicts[key] = axes_dict

        for outcome in outcomes_to_show:
            vmax = -1
            for entry in axes_dicts.values():
                vmax =  max(entry[outcome].images[0].norm.vmax, vmax)
            for entry in axes_dicts.values():
                ax = entry[outcome]
                ax.images[0].set_clim(vmin=0, vmax=vmax)
            del vmax

        return figures, axes_dicts
    else:
        return simple_kde(outcomes, outcomes_to_show, colormap, log, minima,
                          maxima)


def multiple_densities(results,
                       points_in_time=[], 
                       outcomes_to_show=[],
                       group_by = None,
                       grouping_specifiers = None,
                       density=plotting_util.KDE,
                       legend=True,
                       titles={},
                       ylabels={},
                       experiments_to_show=None,
                       plot_type=ENVELOPE,
                       log=False):
    ''' Make an envelope plot with multiple density plots over the run time

    Parameters
    ----------
    results : tupule
              return from :meth:`perform_experiments`.
    points_in_time : list 
                     a list of points in time for which you want to see the 
                     density. At the moment  up to 6 points in time are 
                     supported
    outcomes_to_show : list of str, optional
                       list of outcome of interest you want to plot. If empty, 
                       all outcomes are plotted. **Note**:  just names.
    group_by : str, optional
               name of the column in the cases array to group results by. 
               Alternatively, `index` can be used to use indexing arrays as the
               basis for grouping.
    grouping_specifiers : iterable or dict, optional
                          set of categories to be used as a basis for grouping 
                          by. Grouping_specifiers is only meaningful if 
                          group_by is provided as well. In case of grouping by 
                          index, the grouping specifiers should be in a 
                          dictionary where the key denotes the name of the 
                          group.
    density : {KDE, HIST, VIOLIN, BOXPLOT}, optional
    legend : bool, optional
    titles : dict, optional
             a way for controlling whether each of the axes should have a 
             title. There are three possibilities. If set to None, no title 
             will be shown for any of the axes. If set to an empty dict, 
             the default, the title is identical to the name of the outcome of 
             interest. If you want to override these default names, provide a 
             dict with the outcome of interest as key and the desired title as 
             value. This dict need only contain the outcomes for which you 
             want to use a different title. 
    ylabels : dict, optional
              way for controlling the ylabels. Works identical to titles.
    experiments_to_show : ndarray, optional
                          indices of experiments to show lines for,
                          defaults to None.
    plot_type : {ENVELOPE, ENV_LIN, LINES}, optional
    log : bool, optional

    Returns
    -------
    fig : Figure instance 
    axes : dict
           dict with outcome as key, and axes as value. Density axes' are
           indexed by the outcome followed by _density.
    
    Note
    ---- 
    the current implementation is limited to seven different categories in
    case of group_by, categories, and/or discretesize. This limit is due to the 
    colors specified in COLOR_LIST.
    
    Note
    ----
    the connection patches are for some reason not drawn if log scaling is 
    used for the density plots. This appears to be an issue in matplotlib 
    itself.
    
    '''
    if not outcomes_to_show:
        outcomes_to_show =  results[1].keys()
        outcomes_to_show.remove(TIME)
    elif isinstance(outcomes_to_show, six.string_types):
        outcomes_to_show=[outcomes_to_show]

    axes_dicts = {}
    figures = []
    for outcome_to_show in outcomes_to_show:
        temp_results = copy.deepcopy(results)
        axes_dict = {}
        axes_dicts[outcome_to_show] = axes_dict

        if plot_type != ENV_LIN:
            # standard way of pre processing data
            if experiments_to_show != None:
                experiments, outcomes = temp_results
                experiments = experiments[experiments_to_show]
                new_outcomes = {}
                for key, value in outcomes.items():
                    new_outcomes[key] = value[experiments_to_show]
                temp_results = experiments, new_outcomes

        data = prepare_data(temp_results, [outcome_to_show], group_by, 
                            grouping_specifiers)
        outcomes, outcomes_to_show, time, grouping_labels = data
        del outcomes_to_show

        #start of plotting
        fig = plt.figure()
        figures.append(fig)

        # making of grid
        if not points_in_time:
            raise EMAError("no points in time specified")
        if len(points_in_time) == 1:
            ax_env = plt.subplot2grid((2,3), (0,0), colspan=3)
            ax1 = plt.subplot2grid((2,3), (1,1), sharey=ax_env )
            kde_axes = [ax1]
        elif len(points_in_time) == 2:
            ax_env = plt.subplot2grid((2,2), (0,0), colspan=2)
            ax1 = plt.subplot2grid((2,2), (1,0), sharey=ax_env)
            ax2 = plt.subplot2grid((2,2), (1,1), sharex=ax1, sharey=ax_env)
            kde_axes = [ax1, ax2]
        elif len(points_in_time) == 3:
            ax_env = plt.subplot2grid((2,3), (0,0), colspan=3)
            ax1 = plt.subplot2grid((2,3), (1,0), sharey=ax_env)
            ax2 = plt.subplot2grid((2,3), (1,1), sharex=ax1, sharey=ax_env)
            ax3 = plt.subplot2grid((2,3), (1,2), sharex=ax1, sharey=ax_env)
            kde_axes = [ax1, ax2, ax3]
        elif len(points_in_time) == 4:
            ax_env = plt.subplot2grid((2,4), (0,1), colspan=2)
            ax1 = plt.subplot2grid((2,4), (1,0), sharey=ax_env )
            ax2 = plt.subplot2grid((2,4), (1,1), sharex=ax1, sharey=ax_env)
            ax3 = plt.subplot2grid((2,4), (1,2), sharex=ax1, sharey=ax_env)
            ax4 = plt.subplot2grid((2,4), (1,3), sharex=ax1, sharey=ax_env)
            kde_axes = [ax1, ax2, ax3, ax4]
        elif len(points_in_time) == 5:
            ax_env = plt.subplot2grid((2,5), (0,1), colspan=3)
            ax1 = plt.subplot2grid((2,5), (1,0), sharey=ax_env)
            ax2 = plt.subplot2grid((2,5), (1,1), sharex=ax1, sharey=ax_env)
            ax3 = plt.subplot2grid((2,5), (1,2), sharex=ax1, sharey=ax_env)
            ax4 = plt.subplot2grid((2,5), (1,3), sharex=ax1, sharey=ax_env)
            ax5 = plt.subplot2grid((2,5), (1,4), sharex=ax1, sharey=ax_env)
            kde_axes = [ax1, ax2, ax3, ax4, ax5]
        elif len(points_in_time) == 6:
            ax_env = plt.subplot2grid((2,6), (0,1), colspan=4)
            ax1 = plt.subplot2grid((2,6), (1,0), sharey=ax_env)
            ax2 = plt.subplot2grid((2,6), (1,1), sharex=ax1, sharey=ax_env)
            ax3 = plt.subplot2grid((2,6), (1,2), sharex=ax1, sharey=ax_env)
            ax4 = plt.subplot2grid((2,6), (1,3), sharex=ax1, sharey=ax_env)
            ax5 = plt.subplot2grid((2,6), (1,4), sharex=ax1, sharey=ax_env)
            ax6 = plt.subplot2grid((2,6), (1,5), sharex=ax1, sharey=ax_env)

            kde_axes = [ax1, ax2, ax3, ax4, ax5, ax6, ]
        else:
            raise EMAError("too many points in time provided")

        axes_dict["main plot"] = ax_env
        for n, entry in enumerate(kde_axes):
            axes_dict["density_%s" % n] = entry

            #turn of ticks for all but the first density            
            if n > 0:
                for tl in entry.get_yticklabels():
                    tl.set_visible(False)
            
        # bit of a trick to avoid duplicating code. If no subgroups are 
        # specified, nest the outcomes one step deeper in de dict so the
        # iteration below can proceed normally.
        if not grouping_labels:
            grouping_labels=[""]
            outcomes[""]=outcomes

        for j, key in enumerate(grouping_labels):
            value = outcomes[key][outcome_to_show]

            if plot_type == ENVELOPE:
                plot_envelope(ax_env, j, time, value, fill=False)
            elif plot_type == LINES:
                ax_env.plot(time.T, value.T)
            elif plot_type == ENV_LIN:
                plot_envelope(ax_env, j, time, value, fill=True)
                if experiments_to_show!=None:
                    ax_env.plot(time.T, value[experiments_to_show].T)
                else:
                    ax_env.plot(time.T, value.T)
            ax_env.set_xlim(time[0], time[-1])

            ax_env.set_xlabel(TIME_LABEL)
            do_ylabels(ax_env, ylabels, outcome_to_show)
            do_titles(ax_env, titles, outcome_to_show)

        for i, ax in enumerate(kde_axes):
            time_value = points_in_time[i]
            index = np.where(time==points_in_time[i])[0][0]

            group_density(ax, density, outcomes, outcome_to_show, 
                          grouping_labels, index=index, log=log)

        min_y, max_y = ax_env.get_ylim()
        ax_env.autoscale(enable=False, axis='y')  

        for i, ax in enumerate(kde_axes):
            time_value = points_in_time[i]            

            ax_env.plot([time_value,time_value], 
                        [min_y,max_y], c='k', ls='--')
            con = ConnectionPatch(xyA=(time_value, min_y), 
                                  xyB=(ax.get_xlim()[0],max_y), coordsA="data", 
                                  coordsB="data", axesA=ax_env, axesB=ax)
            ax_env.add_artist(con)

        if legend and group_by:
            lt = PATCH
            alpha = 0.3
            if plot_type==LINES:
                lt = LINE
                alpha = 1
            make_legend(grouping_labels, ax_env, legend_type=lt,alpha=alpha)
    return figures, axes_dicts


'''
Created on 20 May, 2011

This module shows how you can use vensim models directly
instead of coding the model in Python. The underlying case
is the same as used in fluExample

.. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>
                epruyt <e.pruyt (at) tudelft (dot) nl>
'''
from __future__ import (absolute_import, print_function, unicode_literals, 
                        division)

import numpy as np

from ema_workbench import (TimeSeriesOutcome, ScalarOutcome,ema_logging,
                           perform_experiments)
from ema_workbench.em_framework.parameters import Policy, create_parameters

from ema_workbench.connectors.vensim import VensimModel 
from ema_workbench.em_framework.evaluators import MultiprocessingEvaluator

def time_of_max(infected_fraction, time):
    index = np.where(infected_fraction==np.max(infected_fraction))
    timing = time[index][0]
    return timing
    
if __name__ == '__main__':
    ema_logging.log_to_stderr(ema_logging.INFO)

    model = VensimModel("fluCase", wd=r'./models/flu',
                        model_file = r'/FLUvensimV1basecase.vpm')
            
    #outcomes
    model.outcomes = [TimeSeriesOutcome('deceased population region 1'),
                      TimeSeriesOutcome('infected fraction R1'),
                      ScalarOutcome('max infection fraction', 
                                    variable_name='infected fraction R1', 
                                    function=np.max),
                      ScalarOutcome('time of max', 
                                    variable_name=['infected fraction R1', 'TIME'], 
                                    function=time_of_max)]
    
    #create uncertainties based on csv 
    model.uncertainties = create_parameters('./models/flu/flu_uncertainties.csv') 
    
    #add policies
    policies = [Policy('no policy',
                       model_file=r'/FLUvensimV1basecase.vpm'),
                Policy('static policy',
                       model_file=r'/FLUvensimV1static.vpm'),
                Policy('adaptive policy',
                       model_file=r'/FLUvensimV1dynamic.vpm')
                ]
     
     
    with MultiprocessingEvaluator(model, 2) as evaluator:
        results = perform_experiments(model, 1000, policies=policies, 
                                      evaluator==evaluator)




'''

Plotting utility functions

'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)

import copy

import matplotlib as mpl
import matplotlib.cm as cm
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import scipy.stats.kde as kde
import seaborn as sns
import six
from scipy.stats import gaussian_kde, scoreatpercentile

from ..util import EMAError, info, warning

# .. codeauthor:: jhkwakkel <j.h.kwakkel (at) tudelft (dot) nl>

COLOR_LIST = sns.color_palette()
'''Default color list'''
sns.set_palette(COLOR_LIST)

TIGHT = False
'''Parameter controlling whether tight layout from matplotlib should be used'''

TIME = "TIME"
'''Default key for time'''

ENVELOPE = 'envelope'
'''constant for plotting envelopes'''

LINES = 'lines'
'''constant for plotting lines'''

ENV_LIN = "env_lin"
'''constant for plotting envelopes with lines'''

KDE = 'kde'
'''constant for plotting density as a kernel density estimate'''

HIST = 'hist'
'''constant for plotting density as a histogram'''

BOXPLOT = 'boxplot'
'''constant for plotting density as a boxplot'''

VIOLIN = 'violin'
'''constant for plotting density as a violin plot, which combines a
Gaussian density estimate with a boxplot'''

# used for legend
LINE = 'line'
PATCH = 'patch'
SCATTER = 'scatter'

#see http://matplotlib.sourceforge.net/users/customizing.html for details
#mpl.rcParams['savefig.dpi'] = 600
#mpl.rcParams['axes.formatter.limits'] = (-5, 5)
#mpl.rcParams['font.family'] = 'serif'
#mpl.rcParams['font.serif'] = 'Times New Roman'
#mpl.rcParams['font.size'] = 12.0

#==============================================================================
# actual plotting functions
#==============================================================================
def plot_envelope(ax, j, time, value, fill):
    '''
    
    Helper function, responsible for plotting an envelope.
    
    Parameters
    ----------
    ax : axes instance
    j : int
    time : ndarray
    value : ndarray
    fill : bool
    
    
    '''
    
    #plot minima and maxima
    minimum = np.min(value, axis=0)
    maximum = np.max(value, axis=0)

    color = COLOR_LIST[j]
    
    if fill:
#        ax.plot(time, minimum, color=color, alpha=0.3)
#        ax.plot(time, maximum, color=color, alpha=0.3)
        ax.fill_between(time, 
                        minimum,
                        maximum,
                        facecolor=color, 
                        alpha=0.3,
                        )
    else:
        ax.plot(time, minimum, c=color)
        ax.plot(time, maximum, c=color)


def plot_histogram(ax, values, log):
    '''
    
    Helper function, responsible for plotting a histogram

    Parameters
    ----------
    ax : axes instance
    values : ndarray
    log : bool
    
    
    '''
    if isinstance(values, list):
        color = COLOR_LIST[0:len(values)]
    else:
        color=COLOR_LIST[0]
    a = ax.hist(values, 
             bins=11, 
             orientation='horizontal',
             histtype='bar', 
             normed = True,
             color=color,
             log=log)
    if not log:
        ax.set_xticks([0, ax.get_xbound()[1]])
    return a


def plot_kde(ax, values, log):
    '''
    
    Helper function, responsible for plotting a KDE.

    Parameters
    ----------
    ax : axes instance
    values : ndarray
    log : bool
    
    
    '''


    for j, value in enumerate(values):        
        color = COLOR_LIST[j]
        kde_x, kde_y = determine_kde(value)
        ax.plot(kde_x, kde_y, c=color, ms=1, markevery=20)
    
        if log:
            ax.set_xscale('log')
        else:
            ax.set_xticks([int(0), 
                          ax.get_xaxis().
                          get_view_interval()[1]])
            labels =["{0:.2g}".format(0), "{0:.2g}".format(ax.get_xlim()[1])]
            ax.set_xticklabels(labels)


def plot_boxplots(ax, values, log, group_labels=None):
    '''
    helper function for plotting a boxplot
    
    Parameters
    ----------
    ax : axes instance
    value : ndarray
    log : bool
    group_labels : list of str, optional

    
    '''
    
    if log:
        warning("log option ignored for boxplot")
    
    
    ax.boxplot(values)
    if group_labels:
        ax.set_xticklabels(group_labels, rotation='vertical')


def plot_violinplot(ax, value, log, group_labels=None):
    '''
    helper function for plotting violin plots on axes
    
    Parameters
    ----------
    ax : axes instance
    value : ndarray
    log : bool
    group_labels : list of str, optional

    '''
    
    if log:
        warning("log option ignored for violin plot")
    
    pos = range(len(value))
    dist = max(pos)-min(pos)
    _ = min(0.15*max(dist,1.0),0.5)
    for data, p in zip(value,pos):
        if len(data)>0:
            kde = gaussian_kde(data) #calculates the kernel density
            x = np.linspace(np.min(data),np.max(data),250.) # support for violin
            v = kde.evaluate(x) #violin profile (density curve)
            
            scl = 1 / (v.max() / 0.4)
            v = v*scl #scaling the violin to the available space
            ax.fill_betweenx(x,p-v,p+v,facecolor=COLOR_LIST[p],alpha=0.6, lw=1.5)
            
            for percentile in [25, 75]:
                quant = scoreatpercentile(data.ravel(), percentile)
                q_x = kde.evaluate(quant) * scl 
                q_x = [p - q_x, p + q_x]
                ax.plot(q_x, [quant, quant], linestyle=":", c='k')
            med = np.median(data)
            m_x = kde.evaluate(med) * scl 
            m_x = [p - m_x, p + m_x]
            ax.plot(m_x, [med, med], linestyle="--", c='k', lw=1.5)            
        
    if group_labels:
        labels = group_labels[:]
        labels.insert(0, '')
        ax.set_xticklabels(labels, rotation='vertical')

 
def group_density(ax_d, density, outcomes, outcome_to_plot, group_labels, 
                  log=False, index=-1):
    '''
    helper function for plotting densities in case of grouped data
    
    
    Parameters
    ----------
    ax_d : axes instance
    density : {HIST, BOXPLOT, VIOLIN, KDE}
    outcomes :  dict
    outcome_to_plot : str 
    group_labels : list of str
    log : bool, optional
    index : int, optional
    
    Raises
    ------
    EMAError
        if density is unkown
    
    '''
    
    
    
    if density==HIST:
        values = [outcomes[key][outcome_to_plot][:,index] for key in 
                  group_labels]
        plot_histogram(ax_d, values, log)
    elif density==BOXPLOT:
        values = [outcomes[key][outcome_to_plot][:,index] for key in 
                  group_labels]
        plot_boxplots(ax_d, values, log, group_labels)
    elif density==VIOLIN:
        values = [outcomes[key][outcome_to_plot][:,index] for key in 
                  group_labels]
        plot_violinplot(ax_d, values, log, group_labels=group_labels)
    elif density==KDE:
        values = [outcomes[key][outcome_to_plot][:,index] for key in 
                  group_labels]
        plot_kde(ax_d, values, log)
    else:
        raise EMAError("unknown density type: {}".format(density))


def simple_density(density, value, ax_d, ax, log):
    '''
    
    Helper function, responsible for producing a density plot
    
    Parameters
    ----------
    density : {HIST, BOXPLOT, VIOLIN, KDE}
    value : ndarray
    ax_d : axes instance
    ax : axes instance
    log : bool
    
    '''
    
    if density==KDE:
        plot_kde(ax_d, [value[:,-1]], log)
    elif density==HIST:
        plot_histogram(ax_d, value[:,-1], log)
    elif density==BOXPLOT:
        plot_boxplots(ax_d, value[:,-1], log)
    elif density==VIOLIN:
        plot_violinplot(ax_d, [value[:,-1]], log)
    else:
        raise EMAError("unknown density plot type")
        
    ax_d.get_yaxis().set_view_interval(
                 ax.get_yaxis().get_view_interval()[0],
                 ax.get_yaxis().get_view_interval()[1]) 
    ax_d.set_ylim(ymin=ax.get_yaxis().get_view_interval()[0],
              ymax=ax.get_yaxis().get_view_interval()[1])


def simple_kde(outcomes, outcomes_to_show, colormap, log, minima, maxima):
    '''
    
    Helper function for generating a density heatmap over time
    
    Parameters
    ----------
    outcomes : dict
    outcomes_to_show : list of str
    colormap : str
    log : bool
    minima : dict
    maxima : dict
    
    '''

    figure, grid = make_grid(outcomes_to_show)
    axes_dict = {}
    
    # do the plotting
    for i, outcome_to_plot in enumerate(outcomes_to_show):
        ax = figure.add_subplot(grid[i,0])
        axes_dict[outcome_to_plot] = ax
        
        outcome = outcomes[outcome_to_plot]
        
        size_kde = 100
        np.zeros
        kde_over_time = np.zeros(shape=(size_kde, outcome.shape[1]))
        ymin = minima[outcome_to_plot]
        ymax = maxima[outcome_to_plot]
        
        #make kde over time
        for j in range(outcome.shape[1]):
            kde_x = determine_kde(outcome[:, j], size_kde, ymin, ymax)[0]
            if log:
                kde_x = np.log(kde_x+1)
            kde_over_time[:, j] = kde_x
        ax.matshow(kde_over_time, 
                    cmap=cm.__dict__[colormap])
#        a = ax.get_yticklabels()
#        b = len(a)
#        c = ymax-ymin/b
#        d = ymax+c
#        e = np.arange(ymin, d, c)
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.set_xlabel("time")
        ax.set_ylabel(outcome_to_plot)
        
    return figure, axes_dict


def make_legend(categories,
                ax,
                ncol=3,
                legend_type=LINE,
                alpha=1):
    '''
    Helper function responsible for making the legend

    Parameters
    ----------
    categories : str or tuple
                 the categories in the legend
    ax : axes instance 
         the axes with which the legend is associated
    ncol : int
           the number of columns to use
    legend_type : {LINES, SCATTER, PATCH}
                  whether the legend is linked to lines, patches, or scatter 
                  plots
    alpha : float
            the alpha of the artists
    
    '''
    
    some_identifiers = []
    labels = []
    for i, category in enumerate(categories):
        if legend_type == LINE:    
            artist = plt.Line2D([0,1], [0,1], color=COLOR_LIST[i], 
                                alpha=alpha) #TODO
        elif legend_type == SCATTER:
#             marker_obj = mpl.markers.MarkerStyle('o')
#             path = marker_obj.get_path().transformed(
#                              marker_obj.get_transform())
#             artist  = mpl.collections.PathCollection((path,),
#                                         sizes = [20],
#                                         facecolors = COLOR_LIST[i],
#                                         edgecolors = 'k',
#                                         offsets = (0,0)
#                                         )
            # TODO work arround, should be a proper proxyartist for scatter legends
            artist = mpl.lines.Line2D([0],[0], linestyle="none", 
                                      c=COLOR_LIST[i], marker = 'o')

        elif legend_type == PATCH:
            artist = plt.Rectangle((0,0), 1,1, edgecolor=COLOR_LIST[i],
                                   facecolor=COLOR_LIST[i], alpha=alpha)

        some_identifiers.append(artist)
        
        if type(category) == tuple:
            label =  '%.2f - %.2f' % category 
        else:
            label = category
        
        labels.append(str(label))
    
    ax.legend(some_identifiers, labels, ncol=ncol,
                      loc=3, borderaxespad=0.1,
                      mode='expand', bbox_to_anchor=(0., 1.1, 1., .102))


def determine_kde(data, 
                  size_kde=1000,
                  ymin=None,
                  ymax=None):
    '''
    
    Helper function responsible for performing a KDE    
    
    Parameters
    ----------
    data : ndarray
    size_kde : int, optional
    ymin : float, optional
    ymax : float, optional
    
    Returns
    -------
    ndarray
        x values for kde
    ndarray
        y values for kde
        
    ..note:: x and y values are based on rotation as used in density 
             plots for end states.
    
    
    '''
    if not ymin:
        ymin = np.min(data)
    if not ymax:
        ymax = np.max(data)
    
    kde_y = np.linspace(ymin, ymax, size_kde)
    
    try:
        kde_x = kde.gaussian_kde(data)
        kde_x = kde_x.evaluate(kde_y)
#         grid = GridSearchCV(KernelDensity(kernel='gaussian'),
#                             {'bandwidth': np.linspace(ymin, ymax, 20)},
#                             cv=20)
#         grid.fit(data[:, np.newaxis])
#         best_kde = grid.best_estimator_
#         kde_x = np.exp(best_kde.score_samples(kde_y[:, np.newaxis]))
    except Exception as e:
        warning(e)
        kde_x = np.zeros(kde_y.shape)
    
    return kde_x, kde_y
        
def filter_scalar_outcomes(outcomes):
    '''
    Helper function that removes non time series outcomes from all the 
    outcomes.

    Parameters
    ----------
    outcomes : dict
    
    Returns
    -------
    dict
        the filtered outcomes
    
    
    '''
    outcomes_to_remove = []
    for key, value in outcomes.items():
        if len(value.shape) <2:
            outcomes_to_remove.append(key)
            info("%s not shown because it is not time series data" %key)
    [outcomes.pop(entry) for entry in outcomes_to_remove]
    return outcomes


def determine_time_dimension(outcomes):
    '''
    helper function for determining or creating time dimension

    
    Parameters
    ----------
    outcomes : dict
    
    Returns
    -------
    ndarray
    
    
    '''

    time = None
    try:
        time = outcomes['TIME']
        time = time[0, :]
        outcomes.pop('TIME')
    except KeyError:
        values = iter(outcomes.values())
        for value in values:
            if len(value.shape)==2:
                time =  np.arange(0, value.shape[1])
                break
    if time is None:
        info("no time dimension found in results")
    return time, outcomes    


def group_results(experiments, outcomes, group_by, grouping_specifiers,
                  grouping_labels):
    '''
    Helper function that takes the experiments and results and returns a list 
    based on groupings. Each element in the dictionary contains the experiments 
    and results for a particular group, the key is the grouping specifier.
    
    Parameters
    ----------
    experiments : recarray
    outcomes : dict
    group_by : str
               The column in the experiments array to which the grouping 
               specifiers apply. If the name is'index' it is assumed that the 
               grouping specifiers are valid indices for numpy.ndarray.
    grouping_specifiers : iterable
                    An iterable of grouping specifiers. A grouping 
                    specifier is a unique identifier in case of grouping by 
                    categorical uncertainties. It is a tuple in case of 
                    grouping by a parameter uncertainty. In this cose, the code
                    treats the tuples as half open intervals, apart from the 
                    last entry, which is treated as closed on both sides.  
                    In case of 'index', the iterable should be a dictionary 
                    with the name for each group as key and the value being a 
                    valid index for numpy.ndarray.
    
    Returns 
    dict
        A dictionary with the experiments and results for each group, the 
        grouping specifier is used as key
             
    ..note:: In case of grouping by parameter uncertainty, the list of 
             grouping specifiers is sorted. The traversal assumes half open
             intervals, where the upper limit of each interval is open, except 
             for the last interval which is closed.
    
    '''
    groups = {}
    if group_by != 'index':
        column_to_group_by = experiments[group_by]
    
    for label, specifier in zip(grouping_labels, grouping_specifiers):
        if isinstance(specifier, tuple):
            # the grouping is a continuous uncertainty
            lower_limit, upper_limit = specifier
            
            #check whether it is the last grouping specifier
            if grouping_specifiers.index(specifier) ==\
                len(grouping_specifiers)-1:
                #last case
                
                logical = (column_to_group_by>=lower_limit) &\
                           (column_to_group_by<=upper_limit)
            else:
                logical = (column_to_group_by>=lower_limit) &\
                           (column_to_group_by<upper_limit)
        elif group_by =='index':
            # the grouping is based on indices
            logical = specifier
        else:
            # the grouping is an integer or categorical uncertainty
            logical = column_to_group_by==specifier
        
        group_outcomes = {}
        for key, value in outcomes.items():
            value = value[logical]
            group_outcomes[key] = value
        groups[label] = (experiments[logical], group_outcomes)
        
    return groups


def make_continuous_grouping_specifiers(array, nr_of_groups=5):
    '''
    Helper function for discretesizing a continuous array. By default, the 
    array is split into 5 equally wide intervals.
    
    Parameters
    ----------
    array : ndarray
            a 1-d array that is to be turned into discrete intervals.
    nr_of_groups : int, optional
    
    Returns
    -------
    list of tuples 
        list of tuples with the lower and upper bound of the intervals. 
    
    .. note:: this code only produces intervals. :func:`group_results` uses
              these intervals in half-open fashion, apart from the last 
              interval: [a, b), [b,c), [c,d]. That is, both the end point
              and the start point of the range of the continuous array are 
              included.
    
    '''
    
    minimum = np.min(array)
    maximum = np.max(array)
    step = (maximum-minimum)/nr_of_groups
    a = [(minimum+step*x, minimum+step*(x+1)) for x in range(nr_of_groups)]
    assert a[0][0] == minimum
    assert a[-1][1] == maximum
    return a


def prepare_pairs_data(results, 
                        outcomes_to_show=None,
                        group_by=None,
                        grouping_specifiers=None,
                        point_in_time=-1,
                        filter_scalar=True):
    '''
    
    Parameters
    ----------
    results : tuple
    outcomes_to_show : list of str, optional
    group_by : str, optional
    grouping_specifiers : iterable, optional
    point_in_time : int, optional
    filter_scalar : bool, optional
       
    '''
    if isinstance(outcomes_to_show, six.string_types):
        raise EMAError("for pair wise plotting, more than one outcome needs to be provided")
    
    results = copy.deepcopy(results)
    
    outcomes, outcomes_to_show, time, grouping_labels = prepare_data(results, 
                                                        outcomes_to_show,
                                                        group_by,
                                                        grouping_specifiers,
                                                        filter_scalar)

    def filter_outcomes(outcomes, point_in_time):
        new_outcomes = {}
        for key, value in outcomes.items():
            if len(value.shape)==2:
                new_outcomes[key] = value[:, point_in_time]
            else:
                new_outcomes[key] = value
        return new_outcomes
    
    if point_in_time:
        if point_in_time != -1:
            point_in_time = np.where(time==point_in_time)
        
        if group_by:
            new_outcomes = {}
            for key, value in outcomes.items():
                new_outcomes[key] = filter_outcomes(value, point_in_time)
            outcomes = new_outcomes
        else:
            outcomes = filter_outcomes(outcomes, point_in_time)
    return outcomes, outcomes_to_show, grouping_labels 


def prepare_data(results,
                 outcomes_to_show=None,
                 group_by=None,
                 grouping_specifiers=None,
                 filter_scalar=True):
    '''
    
    Parameters
    ----------
    results : tuple
    outcomes_to_show : list of str, optional
    group_by : str, optional
    grouping_specifiers : iterable, optional
    filter_scalar : bool, optional
    
    '''

    #unravel results
    experiments, outcomes = results

    temp_outcomes = {}

    # remove outcomes that are not to be shown
    if outcomes_to_show:
        if isinstance(outcomes_to_show, six.string_types):
            outcomes_to_show = [outcomes_to_show]
            
        for entry in outcomes_to_show:
            temp_outcomes[entry] = copy.deepcopy(outcomes[entry])

    time, outcomes = determine_time_dimension(outcomes)

    # filter the outcomes to exclude scalar values
    if filter_scalar:
        outcomes = filter_scalar_outcomes(outcomes)
    if not outcomes_to_show:
        outcomes_to_show = outcomes.keys()
        
    # group the data if desired
    if group_by:
        if not grouping_specifiers:
            #no grouping specifier, so infer from the data
            if group_by=='index':
                raise EMAError("no grouping specifiers provided while trying to group on index")
            else:
                column_to_group_by = experiments[group_by]
                if column_to_group_by.dtype == np.object:
                    grouping_specifiers = set(column_to_group_by)
                else:
                    grouping_specifiers = make_continuous_grouping_specifiers(column_to_group_by, 
                                                        grouping_specifiers)
            grouping_labels = grouping_specifiers = sorted(grouping_specifiers)
        else:
            if isinstance(grouping_specifiers, six.string_types):
                grouping_specifiers = [grouping_specifiers]
                grouping_labels = grouping_specifiers
            elif isinstance(grouping_specifiers, dict):
                grouping_labels = sorted(grouping_specifiers.keys())
                grouping_specifiers = [grouping_specifiers[key] for key in 
                                       grouping_labels]
            else:
                grouping_labels = grouping_specifiers
                
        
        outcomes = group_results(experiments, outcomes, group_by,\
                                 grouping_specifiers, grouping_labels)
        
        new_outcomes = {}
        for key, value in outcomes.items():
            new_outcomes[key] = value[1]
        outcomes = new_outcomes
    else:
        grouping_labels=[]

    return outcomes, outcomes_to_show, time, grouping_labels


def do_titles(ax, titles, outcome):
    '''
    Helper function for setting the title on an ax
    
    Parameters
    ----------
    ax : axes instance
    titles : dict
             a dict which maps outcome names to titles
    outcome : str
              the outcome plotted in the ax.
    
    '''
    
    if isinstance(titles, dict):
        if not titles:
            ax.set_title(outcome)
        else:
            try:
                ax.set_title(titles[outcome])
            except KeyError:
                warning("key error in do_titles, no title provided for `%s`" % (outcome))
                ax.set_title(outcome)


def do_ylabels(ax, ylabels, outcome):
    '''
    Helper function for setting the y labels on an ax

    Parameters
    ----------
    ax : axes instance
    titles : dict
             a dict which maps outcome names to y labels
    outcome : str
              the outcome plotted in the ax.
    
    '''
    
    if isinstance(ylabels, dict):
        if not ylabels:
            ax.set_ylabel(outcome)
        else:
            try:
                ax.set_ylabel(ylabels[outcome])
            except KeyError:
                warning("key error in do_ylabels, no ylabel provided for `%s`" % (outcome))
                ax.set_ylabel(outcome)    


def make_grid(outcomes_to_show, density=False):
    '''
    Helper function for making the grid that specifies the size and location
    of the various axes. 

    Parameters
    ----------
    outcomes_to_show : list of str
                       the list of outcomes to show
    density: boolean : bool, optional
    
    '''

    
    # make the plotting grid
    if density:
        grid = gridspec.GridSpec(len(outcomes_to_show), 2,
                                 width_ratios = [4, 1])
    else:
        grid = gridspec.GridSpec(len(outcomes_to_show), 1) 
    grid.update(wspace = 0.1,
                hspace = 0.4)
    
    figure = plt.figure()
    return figure, grid




'''
pysd connector
'''
from __future__ import (absolute_import, print_function, division,
                        unicode_literals)
import os

import pysd

from ..em_framework.model import AbstractModel
from ..util import ema_logging
from ema_workbench.util.ema_logging import method_logger
from ema_workbench.em_framework.model import SingleReplication

__all__ = ['PysdModel']

class BasePysdModel(AbstractModel):
    """ Pysd model wrapper

    Parameters
    ----------
    mdl_file: string
        file name of vensim model (e.g. `Teacup.mdl`)
    uncertainties_dict: dictionary
        convenience parameter for constructing normal uncertainties, in the form
        {'Param 1': (lowerbound, upperbound), 'Param 2':(0,1)}
    outcomes_list: list of model variable names
        gets passed to 'return_columns', so can be model variable names
        or their pysafe translations
    working_directory
    name
    """

    def __init__(self, name=None, mdl_file=None):
        if not mdl_file.endswith('.mdl'):
            raise ValueError('model file needs to be a vensim .mdl file')
        if not os.path.isfile(mdl_file):
            raise ValueError('mdl_file not found')
        if name is None:
            name = pysd.utils.make_python_identifier(mdl_file)[0].replace('_','')
        
        super(BasePysdModel, self).__init__(name)
        self.mdl_file = mdl_file
        
        # Todo: replace when pysd adds an attribute for the .py filename
        self.py_model_name = mdl_file.replace('.mdl', '.py')

    @method_logger
    def model_init(self, policy, **kwargs):
        AbstractModel.model_init(self, policy, **kwargs)
        self.model = pysd.read_vensim(self.mdl_file)

    @method_logger
    def run_experiment(self, experiment):
        res = self.model.run(params=experiment,
                             return_columns=self.outcome_variables)
        
        # EMA wants output formatted properly
        output ={col: series.as_matrix() for col, series in res.iteritems()}
        return output


    def reset_model(self):
        """
        Method for reseting the model to its initial state. The default
        implementation only sets the outputs to an empty dict.

        """
        super(BasePysdModel, self).reset_model()
        self.model.reset_state()


class PysdModel(SingleReplication, BasePysdModel):
    pass